

from kabaret.app import plugin
from .flow_view import FlowView

class FlowViewPlugin:

    """
    The default Flow view.

    Will only be installed if no other view
    is registered under the "Flow" view type name.
    """

    @plugin(trylast=True)
    def install_views(session):
        if not session.is_gui():
            return

        type_name = FlowView.view_type_name()
        if not session.has_view_type(type_name):
            session.register_view_type(FlowView)
            session.add_view(type_name)



'''
Generic file resouces repository.

This package provide a way to register some folders under a given name
(:meth:`add_folder`) and later retreive files by their basename and
registered folder name (:meth:`get`).

When registering a folder with the same name as another one, the lookup will
favorize the last one and fall back to the first one.

This way you can extend and/or override the resources available under a given
folder name.

You can deal with any kind of files, but in the case of images there are a couple
helper functions: :meth:`get_pixmap()` and :meth:`get_icon()`

Every search result is cached so disk access is minimized. You don't need to
cache your resources next to their usage code, just can keep calling
``resources.get()`` as much as you want.

You can inspect the available resources with :meth:`get_folder_names()`,
:meth:`list_folder()` and :meth:`list_folder_paths()`

.. note::

    Files starting with an underscore ``_`` are ignored.

'''

import logging
import os
import glob


_PATHS = {}
_CACHE = {}
_PIXCACHE = {}
_ICONCACHE = {}


class ResourcesError(Exception):
    '''
    Raised when the ``folder_name`` is unknown.
    '''


class NotFoundError(ResourcesError):
    '''
    Raised when the the resource was not found.
    Note that this is a :class:`ResourcesError` too.
    '''


def add_folder(name, path):
    '''
    Register a folder to the search path.

    The ``name`` will be used to access files.

    You can use :meth:`get_folder_names` to get a list of all declared names.

    The ``path`` can be a file inside the folder
    to register, or the path of the folder itself.

    If a folder with the same name was previously added, the
    files from the last one overrides the previous one but
    the previous files not available in the last one are
    still accessible.

    In other words: the last call overrides and extends the previous ones.

    Typical resource declaration:

    .. code-block:: python
        :caption: my_studio/icons/emoji/__init__.py

        # We're in a package folder containing some png files
        # Every file will be accessible, and overrides previous
        # declarations of 'icons.emoji' files
        import kabaret.app.resources
        kabaret.app.resources.add_folder('icons.emoji', __file__)

    .. code-block:: python
        :caption: my_studio/gui.py

        # Be sure to have the resource modules imported at least once
        # This is typically done in the module defining your session
        from my_studio.icons import emoji

    Typical resource access:

    .. code-block:: python
        :caption: my_studio/flows/my_dope_flow/__init__.py

        # Use anywhere
        from kabaret.app import resources
        icon = resources.get('icons.emoji', 'sunglasses')

    .. seealso::

        :meth:`get_icon()`
        :meth:`get_pixmap()`

    '''
    global _PATHS, _CACHE

    if not os.path.isdir(path):
        path = os.path.dirname(path)

    if name not in _PATHS:
        _PATHS[name] = []
    _PATHS[name].insert(0, path)

    # Remove cache for this folder name:
    if name in _CACHE:
        _CACHE.pop(name)


def get_folder_names():
    '''
    Returns a list of available folder names.
    '''
    global _PATHS
    return list(_PATHS)


def list_folder(folder_name):
    '''
    Returns a list of resource names available in ``folder_name``.
    '''
    global _PATHS

    EXCLUDE_LIST = ('Thumbs.db',)

    folder_name = folder_name.lower()

    ret = set()
    for path in _PATHS[folder_name]:
        ret.update(
            [
                name.split('.', 1)[0]
                for name in os.listdir(path)
                if not name.startswith('_')
                if not name.startswith('.')
                and name not in EXCLUDE_LIST
            ]
        )
    return sorted(ret)


def list_folder_paths(folder_name):
    '''
    Return a list of list of file_name for ``folder_name``.
    (overrides come first)
    '''
    global _PATHS

    folder_name = folder_name.lower()

    ret = []
    for path in _PATHS[folder_name]:
        ret.append(sorted(
            [
                name.split('.', 1)[0]
                for name in os.listdir(path)
                if not name.startswith('_')
            ]
        ))
    return ret


def get(folder_name, file_name):
    '''
    Returns the file named ``file_name`` in the 
    folder registered as  ``folder_name``.
    '''
    global _PATHS, _CACHE

    # avoid case sensitive look-up in order to have the same
    # behavior in win$ and Linux:
    folder_name = folder_name.lower()
    file_name = file_name.lower()

    cache_key = (folder_name, file_name)
    if cache_key in _CACHE:
        if _CACHE[cache_key]:
            return _CACHE[cache_key]
        else:
            raise NotFoundError("Resource not found: " + repr(cache_key))

    try:
        path_list = _PATHS[folder_name]
    except KeyError:
        raise ResourcesError(
            "Unknown resource folder " +
            repr(folder_name) + ". (Did you install it first?)"
        )

    files = None
    for path in path_list:
        files = glob.glob(
            os.path.join(
                path,
                file_name + ('.' not in file_name and ".*" or "")
            )
        )
        if files:
            break
    if not files:
        _CACHE[cache_key] = None
        raise NotFoundError("Resource not found: " + repr(cache_key))
    if len(files) > 1:
        logging.getLogger('kabaret.resources').warn(
            "Several resources found for: " + repr(cache_key))

    _CACHE[cache_key] = files[0]
    return files[0]


def get_pixmap(folder_name, pixmap_name):
    '''
    Same as :meth:`get()`, but returns a QPixmap
    '''
    global _PIXCACHE

    path = get(folder_name, pixmap_name)
    if path in _PIXCACHE:
        return _PIXCACHE[path]

    from qtpy import QtGui

    ret = QtGui.QPixmap(path)
    _PIXCACHE[path] = ret

    return ret


def get_icon(icon_ref, for_widget=None, disabled_ref=None):
    '''
    Same as :meth:`get_pixmap()`, but returns a QIcon

    If ``icon_ref`` is an **int**,  ``for_widget`` must not be **None**
    and its current QStyle will be used to return
    the QIcon pointed by icon_ref.

    If icon_ref is a 2D tuple, it is used to call
    :meth:`get_pixmap(*icon_ref)`

    If icon_ref is a file path, an icon is created from this file.

    if disable_ref is not None, it must be a 2D tuple suitable for
    :meth:`get_pixmap(*disabled_ref)` and will be used as the "disabled"
    state of the icon.

    Warning: calls with different disabled_ref for the same icon_ref has 
    undefined behavior.

    '''
    if isinstance(icon_ref, int):
        style = for_widget.style()
        return style.standardIcon(icon_ref)

    global _ICONCACHE
    if icon_ref in _ICONCACHE:
        return _ICONCACHE[icon_ref]

    from qtpy import QtGui

    pix = None
    try:
        pix = get_pixmap(*icon_ref)
    except Exception:
        try:
            if os.path.isfile(icon_ref):
                pix = QtGui.QPixmap(icon_ref)
        except TypeError:
            pass

    if pix is None:
        raise NotFoundError('Cannot find icon for %r' % (icon_ref,))

    icon = QtGui.QIcon(pix)

    if disabled_ref is not None:
        pix = get_pixmap(*disabled_ref)
        icon.addPixmap(pix, mode=icon.Mode.Disabled)

    _ICONCACHE[icon_ref] = icon
    return icon



'''
    Value Editor Factory
'''

from .python_value import PythonValueEditor
from .bool_value import BoolValueEditor
from .choice_value import ChoiceValueEditor
from .multichoice_value import MultiChoiceValueEditor
from .textarea import TextAreaEditor
from .label import LabelEditor
from .password_value import PasswordEditor
from .time_value import DateTimeValueEditor, DateValueEditor
from .percent_value import PercentValueEditor
from .thumbnail import ThumbnailEditor
from .path_value import PathValueEditor
from .image_value import ImageValueEditor


_FACTORY = None


class EditorFactory(object):

    def __init__(self):
        super(EditorFactory, self).__init__()
        self._editor_types = []

    def register_editor_type(self, editor_type):
        self._editor_types.append(editor_type)

    def create(self, parent, editor_type_name, options={}):
        # (! reverse the editor type list to have the first reg as default)
        for T in reversed(self._editor_types):
            if T.can_edit(editor_type_name):
                editor = T(parent, options)
                return editor
        raise ValueError(
            'Could not find an editor for %r' % (
                editor_type_name,
            )
        )


def editor_factory():
    global _FACTORY
    if _FACTORY is None:
        _FACTORY = EditorFactory()
        # ! be sure to register this one as the first one as it is universal
        _FACTORY.register_editor_type(PythonValueEditor)
        _FACTORY.register_editor_type(PasswordEditor)
        _FACTORY.register_editor_type(BoolValueEditor)
        _FACTORY.register_editor_type(ChoiceValueEditor)
        _FACTORY.register_editor_type(MultiChoiceValueEditor)
        _FACTORY.register_editor_type(TextAreaEditor)
        _FACTORY.register_editor_type(LabelEditor)
        _FACTORY.register_editor_type(DateTimeValueEditor)
        _FACTORY.register_editor_type(DateValueEditor)
        _FACTORY.register_editor_type(PercentValueEditor)
        _FACTORY.register_editor_type(ThumbnailEditor)
        _FACTORY.register_editor_type(PathValueEditor)
        _FACTORY.register_editor_type(ImageValueEditor)

    return _FACTORY



'''

    kabaret.flow.action

    Defines the Action class, an object on which UI commands can call
    the run() method to execute some code.

    The return value of run() dictates some UI behavior
    (see Action.run() doc).

    Defines the DropAction class, an object used to handle GUI drops
    on the ConnectAction parent.
    (NB: it is not a subclasse of Action and wont show up as a button)

'''

from .object import Object
from .relations import SessionParam, Parent


class Action(Object):

    ICON = 'action'

    # does this action shows in parent's inline GUI ?
    SHOW_IN_PARENT_INLINE = True
    # does this action shows in parent's detailed GUI ?
    SHOW_IN_PARENT_DETAILS = True

    message = SessionParam().ui(editor='label', wrap=True).ui(editable=False, label='')

    @classmethod
    def get_result(
            cls, close=None, refresh=None,
            goto=None, goto_target=None, goto_target_type=None, 
            next_action=None
    ):
        '''
        The value returned by run(button) will be inspected by callers to
        decide how they should react.
        This helper class method will generate a result matching its args:
            close: (bool)
                Prevent the action dialog to be closed when set to False
            refresh: (bool)
                Force a reload of the parent page avec dialog close
                (This should not be needed anymore thanks to touch() calls...)
            goto: (oid)
                Force the caller to load the given oid page.
            goto_target: (string or "_NEW_")
                The string identifer of the view targeted by the goto directive.
                The default targeted view is the current one or the one that created
                the action dialog. Spcecifying a target can be usefull to alter
                another view. If the specified target is not found, a new view
                will be created with this identifier so that further use of the
                action will affect this new view.
                The special identifier "_NEW_" can be used to force the creation
                of a new view.
            goto_target_type: (string)
                The type of the view targeted by the goto directive.
                Default is 'Flow'
                An easy way to find out the available type names is to use
                a obviously invalid value like "???" and look for the error
                message: it will contain a list of available type names.
            next_action: (oid)
                Force the dialog to not close but load the ui for the action
                with the given oid.
                This is the way to implement "Wizard Pages"

        '''
        ret = {}
        if close is not None:
            ret['close'] = close and True or False
        if refresh is not None:
            ret['refresh'] = refresh and True or False
        if goto is not None:
            ret['goto'] = goto
        if goto_target is not None:
            ret['goto_target'] = goto_target
        if goto_target_type is not None:
            ret['goto_target_type'] = goto_target_type

        if next_action is not None:
            ret['next_action'] = next_action

        return ret

    def allow_context(self, context):
        '''
        Override this to control where the action will be accessible.
        Return True to allow the action in the given context, False to disallow it,
        and None to fall back to legacy behavior (SHOW_IN_PARENT_* class attributes)

        Default is to return None.

        :param context: string. the context fetching the actions to show.
        Usually the type name of the View, optionally followed by sub-categories.
        The built-in Flow view will send:
            Flow.details    the action's parent details representation
            Flow.inline     the action's parent inline representation
            Flow.indline>>> the inline representation of  (the number of '>' depicts the depth of the inline
            action)
        :return: True/False/None
        '''
        if context is None:
            import logging
            logger = logging.getLogger(__name__)
            logger.warning('Action.allow_context() with context=None ! {}'.format(self.oid()))
        return None

    def needs_dialog(self):
        '''
        May be overriden by subclasses to return False if the action
        does not need to show a dialog.
        Action without dialog are called with run(button=None), plus the 'goto'
        and 'next_action' fields of the returned value do the same.

        NB: If the action cannot run, it should show a dialog with a
        desciption why in self.message (override get_button() to do so).
        '''
        return True

    def get_buttons(self):
        '''
        Returns a list of string suitable as the 'button' argument for a
        call to self.run().

        NB: If the action cannot run, you should set a desciption why
        in self.message, return ['Cancel'] and handle that in run().
        This is far better than returning False from needs_dialog() or let
        the run() method decide to do nothing since the user will not have
        a clear feedback showing that nothing happened...
        '''
        return ['Run']

    def run(self, button):
        '''
        The return value should be None or the return value of
        a self.get_result() call.
        GUI will inspect the returned value and act upon it.
        '''
        self.message.set('This is an abstract action and it cannot be used.')
        return self.get_result(close=False)


class ConnectAction(Object):
    '''
    Subclasses will want to overwrite those methods:
        - accept_label: 
            return the label to show in GUI if the objects and urls are
            acceptable, None otherwise.
        - run: 
            to do the job.

    The run method is guaranted to be called only if accept_label did not return None.
    '''

    def accept_label(self, objects, urls):
        return 'Drop %i Oject(s)/File(s) here' % (len(objects) + len(urls),)

    def run(self, objects, urls):
        raise NotImplementedError()


class ChoiceValueSetAction(Action):

    SHOW_IN_PARENT_INLINE = False
    VALUE_TO_SET = None

    _choice_value = Parent()

    def needs_dialog(self):
        return False

    def run(self, button):
        self._choice_value.set(self.VALUE_TO_SET)


class ChoiceValueSelectAction(Action):

    SHOW_IN_PARENT_INLINE = False

    _choice_value = Parent()

    def get_buttons(self):
        return self._choice_value.choices()

    def run(self, button):
        self._choice_value.set(button)




import random

from kabaret import flow

from kabaret.app import resources

class ResourceIconObject(flow.Object):

    ICON = ('icons.status', 'DONE')

    doc = flow.Label(
        'The icon must be a 2d tuple leading to an image '
        'resource (see the Icons section at the root of '
        'this project).'
    )

class MapExample(flow.DynamicMap):

    @classmethod
    def mapped_type(cls):
        return flow.Object

    def mapped_names(self, page_num=0, page_size=None):
        return ['Item{}'.format(i) for i in range(10)]

    def columns(self):
        return ['Name', 'Col1', 'Col2']

    def _fill_row_cells(self, row, item):
        row.update(dict(
            Name=item.name(),
            Col1=item.name()+'@col1',
            Col2=item.name()[-1],
        ))

    def _fill_row_style(self, style, item, row):
        statuses = resources.list_folder('icons.gui')
        status = random.choice(statuses)
        bg = '#FF8800' # may also be like (255, 127, 0)
        fg = '#008888'
        if int(row['Col2'])%3:
            bg = '#00FF00'
            fg = '#004400'
        status2 = random.choice(statuses)
        style.update({
            'icon':('icons.gui', status),
            'Col1_foreground_color':fg,
            'Col1_background_color':bg,
            'Col2_icon':('icons.gui', status2),
        })

class AlteredIconValue(flow.values.Value):

    EDITABLE = True

    def _fill_ui(self, ui):
        status = random.choice(
            resources.list_folder('icons.status')
        )
        ui['placeholder'] = 'The "icon" and "editable" properties change at each Refresh...'
        ui['icon'] = ('icons.status', status)
        ui['editable'] = self.EDITABLE
        self.EDITABLE = not self.EDITABLE


class ExpandedObject(flow.Object):

    message = flow.Label(
        """<h2>Please Don't fold me !</h2>
        <h3>The "expanded" property override the default expanded state, set it to <b>True</b> to expand automatically 
        an Object when it appear for the first time<h3>"""
    )


class NotExpandableObject(flow.Object):

    message = flow.Label(
        """
        <h2>Thanks for coming but there is nothing to see here :/</h2>
        <h2>Have a nice day !</h2>
        """
    )
    parent = flow.Parent().ui("Go Back")

class HiddenNavBarAction(flow.Action):

    sub_object = flow.Child(flow.Object)

    def allow_context(self, context):
        return 'detail' in context

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set(
            "<html>"
            "<p>This Action dialog doesn't have a navigation bar.</p>"
            "<p>Sub object may still show it though...</p>"
            "</html>"
        )
        return ['Nice']

    def run(self, button):
        return None

    def _fill_ui(self, ui):
        ui['navigation_bar'] = False

class HiddenNavBar(flow.Object):

    doc = flow.Label(
        """
        <h2>Hidden Navigation Bar</h2>

        <p>
        This page does not show the navigation bar 
        which can be usefull to lighten the GUI in Action dialog
        and custom pages for example...
        </p>
        <p>
        You can use 
        <pre>   Relation.ui(navigation_bar=False)</pre>
        to control it,
        and you can also set it in 
        <pre>   Object._fill_ui()</pre>
        </p>
        <p><font color=cyan>
        You can use the <b>Option</b> menu to 
        <b>toggle</b> the navigation bar back on.
        </p></font>
        """
    )

    parent_page = flow.Parent()

    hidden_nav_bar_action = flow.Child(HiddenNavBarAction)

class UIConfigGroup(flow.Object):

    doc = flow.Label(
        '''
        <hr><h2>There are a few options to configure the look and
        interface for your Objects.</h2>
        '''
    )

    icon_doc = flow.Label(
        '''
        <hr><h3>The <b>ICON</b> class attribute will associate an icon with your
        Object or Action class. This icon will appear in most place in the GUI
        where your Object appears.</h2>
        '''
    )
    object_with_icon = flow.Child(ResourceIconObject)

    relation_doc = flow.Label(
        '''
        <hr><h3>When defining a relation to another object, you can specify 
        some ui options</h3>
        Those are:
        <pre>
            icon, editor_type, editable, label, hidden
        </pre>
        The icon will override the one defined in the Object class:<br>
        <pre>
            object_with_icon = flow.Child(MyObject).ui(another)
        </pre>
        Note that you can chain the call to Relation.ui():
        <pre>
            my_object = flow.Child(MyObject).ui(
                icon=my_icon
            ).ui(
                label='This Label Is Overriden'
            )
                
        </pre>
        </h3>
        '''
    )
    object_with_overridden_icon = flow.Child(ResourceIconObject).ui(
        icon=('icons.status', 'WIP')
    )

    fill_doc = flow.Label(
        '''
        <hr><h3>
        Sometime you need to have an icon or other ui properties changing
        over time or per instance.
        </h3>
        To achieve this, you can override the `Object._fill_ui(ui)` method
        and alter the default ui passed to it.<br>
        The ui argument is a dict containing all ui properties set on the 
        relation. Just change a value to after the ui (Beware: changes will only
        show after a GUI refresh or navigation !). 
        '''
    )
    altered_icon = flow.Param('', AlteredIconValue)

    map_doc = flow.Label(
        '''
        <hr><h3>The Map items can be styles with icons and colors. There are many 
        option, please refer to the docstring of `DynamicMap.rows()` method.</h3>
        (refresh to see new random colors/icons)
        '''
    )
    map_example = flow.Child(MapExample)

    expanded_example = flow.Child(ExpandedObject).ui(expanded=True)

    expandable_doc = flow.Label(
        "<hr><h3>You can set the `expandable` property to forbid the expanding of an Object, "
        "forcing the user to display it in a new page (by double clicking)</h3>"
    )
    not_expandable = flow.Child(NotExpandableObject).ui(expandable=False)

    hidden_nav_bar = flow.Child(HiddenNavBar).ui(
        navigation_bar=False,
        expandable=False,
    )



'''

Dependency Injection / Invertion of Control for Flow tree
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

In some situations, you need to change the type of an Object in
a flow that is defined by some code you can't (or don't want to) modify.

*kabaret* let you do that on relations that have been flagged as 
`injectable()` by their author.

All `Child` relations (`Param`, `Computed`, `Connection`, ...) can be
injected, as well as the `Map.mapped_type(cls)` method. Here is how
to proceed:

Child Relations
^^^^^^^^^^^^^^^

1) The lib flow must flag the relation as injectable:

.. code-block:: python

    # in module awesome_flow_lib:

    class LibObject(flow.Object):
        # `inherit_default` arg is optional, defaults to True
        lib_child = flow.Child(LibChild).injectable(inherit_default=True)

2) Define your custom Object to use, then provide it in an ancestor of `LibObject`
that inherits `kabaret.flow.InjectionProvider` or just implements a 
`_injection_provider(slot_name, default_type)` classmethod:

.. code-block:: python

    from kabaret import flow
    import awesome_flow_lib

    class MyChild(awesome_flow_lib.LibChild):
        additionnal_param = flow.Param('Tadaaa !')

    class MyProject(flow.Object, flow.InjectionProvider):

        lib_object = flow.Child(awesome_flow_lib.LibOject)

        @classmethod
        def _injection_provider(cls, slot_name, default_type):
            if defalut_type is LibChild:
                return MyChild


Mapped Type
^^^^^^^^^^^

1) The lib flow must use `flow.injection.injectable()` in the `mapped_type(cls)` classmethod
of the injectable `Map`.

.. code-block:: python

    # in module awsome_flow_lib:

    class LibMap(flow.Map):

        @classmethod
        def mapped_type(cls):
            return flow.injection.injectable(
                LibMappedObject,
                # Optional, defaults to True:
                inherit_default=True,
            )

2) Define your custom Object to use, then provide it in an ancestor of `LibMap`
that inherits `kabaret.flow.InjectionProvider` or just implements a 
`_injection_provider(slot_name, default_type)` classmethod:

.. code-block:: python

    from kabaret import flow
    import awesome_flow_lib

    class MyMappedObject(awesome_flow_lib.LibMappedObject):
        additionnal_param = flow.Param('Tadaaa !')

    class MyProject(flow.Object, flow.InjectionProvider):

        lib_map = flow.Child(awesome_flow_lib.LibMap)

        @classmethod
        def _injection_provider(cls, slot_name, default_type):
            if slot_name == 'awesome_flow_lib.LibMappedObject':
                return MyMappedObject


'''
import logging
logger = logging.getLogger(__name__)

class InjectionSlotError(ValueError):
    pass

class InjectionInheritError(TypeError):
    def __init__(self, slot_name, injected_type, default_type):
        super(InjectionInheritError, self).__init__(
            "The injected type for slot '{}' must inherit '{}', got '{}'".format(
                slot_name, injected_type, default_type
            )
        )

class Injection(object):

    def __init__(self):
        super(Injection, self).__init__()
        self._inherit_defaults = {} # slot_name: bool
        self._cache = {}            # (context_name,slot_name): injected_type

    def _get_full_qualified_name(self, T):
        return '%s.%s' % (T.__module__, T.__name__)

    def get_context_name(self, o):
        project = o.root().project()
        return self._get_full_qualified_name(project.__class__)

    def get_slot_name(self, default_type):
        return self._get_full_qualified_name(default_type)

    def injectable(self, default_type, inherit_default=True):
        slot_name = self.get_slot_name(default_type)
        logger.debug("Declaring injection '{}'".format(slot_name))
        try:
            current_inherit_default = self._inherit_defaults[slot_name]
        except KeyError:
            self._inherit_defaults[slot_name] = inherit_default
        else:
            if current_inherit_default != inherit_default:
                raise InjectionSlotError(
                    "An incompatible injection already exists for '{}': "
                    "it has inherit_default=={:r}".format(
                        slot.inherit_default
                    )
                )
        return default_type

    def resolve(self, default_type, parent):
        slot_name = self.get_slot_name(default_type)
        try:
            inherit_default = self._inherit_defaults[slot_name]
        except KeyError:
            # Not slot defined for this type:
            logger.debug(
                "No injection declared for '{}', using default type.".format(
                    slot_name
                )
            )
            return default_type

        cache_key = (self.get_context_name(parent), slot_name)
        try:
            return self._cache[cache_key]
        except KeyError:
            logger.debug(
                "Resolving injection for '{}'.".format(
                    slot_name
                )
            )
        else:
            logger.debug(
                "Using cached injection for '{}'.".format(
                    slot_name
                )
            )

        injected_type = self._provide(slot_name, default_type, parent)
        if injected_type is default_type:
            logger.debug(
                "Injection for '{}' resolved to default type.".format(
                    slot_name
                )
            )
        elif inherit_default:
            if not issubclass(injected_type, default_type):
                raise InjectionInheritError(slot_name, injected_type, default_type)
        self._cache[slot_name] = injected_type
        return injected_type

    def _provide(self, slot_name, default_type, parent):
        '''
        Walk up all flow ancestors, resolving the injection at each step.
        Returns the last injected type.
        '''
        try:
            injected_type = parent._injection_provider(slot_name, default_type)
        except (AttributeError, TypeError):
            # The parent doesnt define `_injection_provider`
            # or does not define it as a callable:
            injected_type = default_type
        else:
            if injected_type is None:
                # The parent defines the `_injection_provider` method
                # but it didn't handle this injection
                injected_type = default_type

        parent = parent._mng.parent
        if parent is not None:
            injected_type = self._provide(slot_name, injected_type, parent)

        return injected_type
    
injection = Injection()

class InjectionProvider(object):
    '''
    This is a Mixin class used to implement injection
    provider in you flows.

    It's not mandatory to use it, you can simply add
    a _injection_provider(slot_name, default_type) class
    method in you flow Object(s). But this implementation
    prints out all requests so it's a nice way to see
    all the injectable slot_name before implementing
    your provider...
    '''
    def __init__(self):
        super(InjectionProvider, self).__init__()
    
    @classmethod
    def _injection_provider(cls, slot_name, default_type):
        logger.info(
            (
                'Default Injection Provider requested:\n'
                '  provider: {}\n'
                '  slot name: {}\n'
                '  deafult_type: {}\n'
            ).format(cls, slot_name, default_type)
        )



import sys

from qtpy import QtCore

from kabaret.app.ui import gui

from kabaret.app.actors.flow import Flow

_has_button_page_and_script_view = False
try:
    from kabaret.flow_button_pages import ButtonHomeRoot
    from kabaret.script_view.script_view import ScriptView
except ImportError:
    pass
else:
    _has_button_page_and_script_view = True


class MyGUISession(gui.KabaretStandaloneGUISession):

    def _create_actors(self):
        if _has_button_page_and_script_view:
            Flow(self, CustomHomeRootType=ButtonHomeRoot)
        else:
            return super()._create_actors()

    def register_view_types(self):
        if _has_button_page_and_script_view:
            type_name = self.register_view_type(ScriptView)
            self.add_view(type_name, hidden=False, area=QtCore.Qt.RightDockWidgetArea)

            self.main_window_manager.main_window.resize(3000, 2000)

        return super().register_view_types()


def start_kabaret(args):
    session_name="dev_studio"
    host = 'redis.test.com'
    port="6379"
    cluster_name="DEV_STUDIO"
    db_index="0"
    password='SUPER_SECRET_PASSWORD'
    read_replica_host = host # set None to deactivate read replica
    read_replica_port = '6378'

    debug = True
    layout_mgr = False
    layout_autosave = False
    layout_savepath = None

    if args:
        # Use -h in command line to see the flags!
        (
            session_name,
            host, port, cluster_name,
            db_index, password, debug,
            read_replica_host, read_replica_port,
            layout_mgr, layout_autosave, layout_savepath,
            remaining_args
        ) = MyGUISession.parse_command_line_args(args)
    print('Will connect using:', (
        session_name,
        host, port, cluster_name,
        db_index, password, debug,
        read_replica_host, read_replica_port,
    ))

    session = MyGUISession(
        session_name=session_name, 
        debug=debug, 
        layout_mgr=layout_mgr, layout_autosave=layout_autosave, layout_savepath=layout_savepath,
    )
    session.cmds.Cluster.connect(
        host=host,
        port=port,
        cluster_name=cluster_name,
        db_index=db_index,
        password=password,
        read_replica_host=read_replica_host,
        read_replica_port=read_replica_port,
    )
    session.start()
    session.close()


if __name__ == "__main__":
    start_kabaret(sys.argv[1:])



from kabaret import flow


#
#--- Lib Code
#
# Let's pretend this code is in a pip install package we can't 
# modify.
# 

class DeptChoiceValue(flow.values.MultiChoiceValue):

    CHOICES = ('Preprod', 'Prod', 'Edit')


class User(flow.Object):
    '''
    Base class for User in the Lib.
    Use the slot name <libmodule>.User to inject
    a custom User subclass.

    This User has a dept Param with a default 
    choice list.
    Use the slot name <libmodule>.DeptChoiceValue to
    inject a custom flow.values.ChoiceValue.
    '''
    @classmethod
    def get_source_display(cls, oid):
        return oid.rsplit('/', 1)[-1]
    
    # By default, the injection must be a subclass of the 
    # original. With inherit_default=False this restriction
    # is disabled:
    depts = flow.Param([], DeptChoiceValue).injectable(inherit_default=False)
    email = flow.Param('')

class AddUserAction(flow.Action):

    _users = flow.Parent()

    def needs_dialog(self): return False
    def run(self, button):
        user = self._users.add('User%03d'%len(self._users))
        self._users.touch()

class ClearUsersAction(flow.Action):

    _users = flow.Parent()

    def needs_dialog(self): return False
    def run(self, button):
        self._users.clear()
        self._users.touch()

class Users(flow.Map):
    '''
    A collection of Users.

    You can inject your custom User type using the slot
    <libmodule>.User

    You can inject you add_user and clear_users Action using
    the slot names:
    - <libmodule>.AddUserAction
    - <libmodule>.ClearUsersAction
    '''
    add_user = flow.Child(AddUserAction).injectable(inherit_default=False)
    clear_users = flow.Child(ClearUsersAction).injectable(inherit_default=False)

    @classmethod
    def mapped_type(cls):
        return flow.injection.injectable(User)
    
    def columns(self):
        return ['Name', 'User Type', 'Department(s)']
    
    def _fill_row_cells(self, row, item):
        row['Name'] = item.name()
        row['User Type'] = item.__class__.__name__
        row['Department(s)'] = ', '.join(item.depts.get())
        
class Task(flow.Object):

    start_date = flow.Param(0).ui(editor='date')
    nb_days = flow.IntParam(2)
    assigned = flow.Connection(User)


#
#--- Customization code
#

class MyDepts(flow.values.MultiChoiceValue):
    '''
    This will be injected in the lib to customize the
    list of departments.
    '''
    CHOICES = ('Mod', 'Rig', 'Surf', 'Anim', 'Lighting', 'Comp')

class MyUser(User):
    '''
    Our user has a squad param and a chapter_head boolean param.
    We will inject this User subclass in the lib code.
    '''
    squad = flow.Param('')
    chapter_head = flow.BoolParam(False)

class Demo(flow.Object, flow.InjectionProvider):
    '''
    This would be the Projet class.
    It uses the public Object from the lib (Task and Users),
    but it injects a couple type that the lib will use
    instead of the default ones.

    Note that this Object inherits flow.InjectionProvider
    This is not mandatory, you just need to implement
    the _injection_provider() class method. But the
    default implementation in flow.InjectionProvider
    will print injection request which is usefull
    to see all slot_names before starting to implement
    your provider...
    '''
    task1 = flow.Child(Task)
    task2 = flow.Child(Task)

    users = flow.Child(Users).ui(expanded=True)

    @classmethod
    def _injection_provider(cls, slot_name, default_type):
        # In this demo, the lib module name is the current 
        # one because it's not really a lib.
        # We're using __name__ but in real cases you'll
        # need to write the full qualified name like:
        # 'awesome_flow_users_lib.users.DeptChoiceValue'
        if slot_name == __name__+'.DeptChoiceValue':
            return MyDepts

        elif slot_name == __name__+'.User':
            return MyUser

#
#--- Simple Demo
#
class LibChild(flow.Object):
    lib_param = flow.Param('This is a lib Param').ui(editable=False)

class LibObject(flow.Object):
    child = flow.Child(LibChild).injectable()

class MyChild(LibChild):
    my_param = flow.Param('This is my additionnal Param').ui(editable=False)

class SimpleDemo(flow.Object, flow.InjectionProvider):
    doc = flow.Label(
        'This is a simple demonstration of Child and Map injection.\n'
        'The "some_lib_object" here contain a "child" object with 2 Params. '
        'One of those param was defined outside the lib.\n'
        '(Read the code as a reference.)'
    )
    some_lib_object = flow.Child(LibObject)

    @classmethod
    def _injection_provider(cls, slot_name, default_type):
        if default_type is LibChild:
            return MyChild

class InjectionGroup(flow.Object):

    demo = flow.Child(Demo).ui(extended=True)
    simple_demo = flow.Child(SimpleDemo)

    intro = flow.Label(
        '''
        <html><body>
        <font size=+2>

        <H1>Dependency Injection / Inversion of Control</H1>
        <H2>Why ?</H2>
        <p>
        Hi-level Objects often need to relate each others.
        </p>
        <p>
        When building a self contained flow, this is not an issue.
        But while building flow libraries, you face a situation where
        you have to deliver Objects for every related high level 
        concepts so that they can interact thru their relations.
        </p>
        <p>
        The situation is also bad for the library users who will need
        to adopt all the concepts of the library because they can't
        change the relation between them.
        </p>
        <p>
        As an example, let's consider you want to build a Task 
        system:<br>
        You use a `Map` to manage a list of `Task`.<br>
        A `Task` has `start_date` and `duration`, as well as a 
        `assigned_user` which is a `Connection` to a `User`.<br>
        At this point you need to define what is a `User` and the 
        interface it needs to interact with your `Task`. So you had 
        a `Team` map containing `User` items...<br>
        You Task system now contains Tasks and Users and in order to
        use it in a pipeline flow one needs to adopt your Task system
        as well as your User system. <b>This sucks !!!</b>
        </p>
        <p>
        On this example, an <b>Inversion of Control</b> is to give
        the pipeline flow the control over the relation used by the Task
        lib flow.<br>
        A common way to achieve this is to <b>inject</b> the 
        <b>dependencies</b> inside the Task flow lib from the pipeline flow
        instead of letting the Task flow lib define them.
        </p>
        <H3>Why is it a problem ?</H3>
        <p>
        A flow Object dependencies are defined by the `Child` and `Map`
        contained in all its children. (`Param` are `Child`, but `Parent` 
        and `Relative` are special and dont apply here).<br>
        More precisely, they are defined by the type of Objects
        each `Relation` relate to and the type of item mapped to each `Map`.<br>
        The classic way to change those (in order to change the dependencies) is
        by inheritance: you inherit the relation owner and override the relation 
        by defining a new one with the same name.<br>
        The probleme is that you now need to use this new class instead of the one 
        containing the Relation you altered. So you also need to inherit its
        parent's class. And this goes up to your Project class !<br>
        <b>This sucks !!!</b><br>
        But keep your pants on and read on, we have a
        solution for this situation: <b>Dependency Injection</b>
        </p>
        <p>
        Also, it's worth specifying that dependency injection also makes writing 
        test code extremely easier as it lets you replace/mock-up part of the
        system to isolate the component you want to test.
        </p>
        <H2>How ?</H2>
        <p>
        So we want to "override" the <b>related type</b> of some Relations, and
        the <b>mapped type</b> of some Maps.<br>
        We want to do it with different types on different flow, be them pipeline
        flows or lib flows.<br>
        And as alwawy, we want to do from within the flow itsef so that everything
        is well tight together.<br>
        Also, as a lib flow developer, we want to specify which types can be 
        <b>injected</b> ("overidden") in my lib Objects.
        </p>
        <p>
        The first step is specifying that a Relation supports injection. This is 
        done by calling `injectable()` on it.<br>
        In this example, the `my_studio.flow.lib.foo` lib flow defines a `Foo` 
        Object that contains a `FooChild` Object. Let's say you want your lib users
        to be able to inject their Object intead of using your 'FooChild':
        <pre>
            # my_studio/flow/lib/foo.py`

            class Foo(flow.Object):    
                my_child = flow.Child(FooChild).injectable()

        </pre>
        The second step is to use the foo lib and inject a new type instead of
        `FooChild`.<br>
        This is done by defining an `_injection_provider()` method in any parent
        of the `Foo` Object.<br>
        In this example, we inject a `FooChild` subclass that extends it with
        the `is_awesome` Param.
        <pre>
            from my_studio.flow.lib.foo import Foo, FooChild

            class MySuperFooChild(FooChild):
                is_super = flow.BoolParam(True)

            class Project(flow.Object, flow.InjectionProvider):

                foo = flow.Child(Foo)

                def _injection_provider(self, slot_name, default_type):
                    if default_type is FooChild:
                        return MySuperFooChild
        </pre>
        With this set up, your `Project` pipeline flow is using the `foo` lib
        flow with an extended FooChild.
        <b>This rocks !!!</b>
        </p>

        <H2>Demo</H2>
        <p>
        In the "Demo" group above here you can see a couple of Tasks and a 
        Team of Users.<br>
        Each Task can be assigned a User by drag'n'drop.<br>
        </p>
        <p>
        The Task type is defined in a library. It uses a default User type 
        with a 'dept' param with 'Preprod', 'Prod' and 'Edit' multiple choices.
        </p>
        <p>
        We want to use this library but we have more departments for our users,
        and we also have squad and chapters.<br>
        We define a new User class by inheriting the lib's one and extending it
        with more params. This User type in then injected into the lib.<br>
        We also Inject a custom DeptChoiceValue into the lib to have our own
        department choices.
        </p>
        <H2>Notes</H2>
        <p>
        This demo can't showcase it but injections are tied to the current
        flow root so that they can be different per project. This means you
        can use the same flow lib injected differently in several pipelines.<br>
        On the other hand, the injections are resolved only once per pipeline
        and your providers will never be called twice on the same `slot_name`. 
        This means that all instances of a Project will use the same 
        injections. This is a feature and not a restriction: a [work/data]flow 
        must be consistent in all its occurences in order to provide all the
        intended benefits. 
        </p>
        </font>
        </body></html>
   '''
    )



from kabaret import flow

class NoDialogAction(flow.Action):
    def needs_dialog(self): return False

class ObjectWithSubmenus(flow.Object):

    with flow.group('Group A'):
        group_a_1 = flow.Child(flow.Action)
        group_a_2 = flow.Child(NoDialogAction)

    with flow.group('Group A/SubGroup A'):
        group_aa_1 = flow.Child(flow.Action)
        group_aa_2 = flow.Child(NoDialogAction)

    with flow.group('Group A/SubGroup B'):
        group_ab_1 = flow.Child(flow.Action)
        group_ab_2 = flow.Child(NoDialogAction)

    with flow.group('Group B'):
        group_b_1 = flow.Child(flow.Action)
        group_b_2 = flow.Child(NoDialogAction)

    with flow.group('Group B/SubGroup A'):
        group_ba_1 = flow.Child(flow.Action)
        group_ba_2 = flow.Child(NoDialogAction)
    
class AddItemAction(flow.Action):
    _map = flow.Parent()
    def needs_dialog(self): return False
    def run(self, button): self._map.do_add_item()

class MapWithSubmenus(flow.Map):

    add_item = flow.Child(AddItemAction)
    with flow.group('GroupA'):
        action_A1 = flow.Child(flow.Action)
        action_A2 = flow.Child(flow.Action)
    with flow.group('GroupA/SubA'):
        action_AA1 = flow.Child(flow.Action)
        action_AA2 = flow.Child(flow.Action)
    with flow.group('GroupA/SubB'):
        action_AB1 = flow.Child(flow.Action)
        action_AB2 = flow.Child(flow.Action)
    with flow.group('GroupB'):
        action_B1 = flow.Child(flow.Action)
        action_B2 = flow.Child(flow.Action)
    with flow.group('GroupB/SubA'):
        action_BA1 = flow.Child(flow.Action)
        action_BA2 = flow.Child(flow.Action)
    with flow.group('GroupB/SubB'):
        action_BB1 = flow.Child(flow.Action)
        action_BB2 = flow.Child(flow.Action)
        
    @classmethod
    def mapped_type(cls):
        return ObjectWithSubmenus
    
    def do_add_item(self):
        name = 'Item{:03}'.format(len(self))
        self.add(name)
        self.touch()

class ActionSubmenusDemo(flow.Object):

    doc = flow.Label(
        'See code in "{}" to learn how to activate Action Submenu like '
        'here.'.format(__name__)
        +'<pre>'
        '''
TL;DR:
    object_with_submenus = flow.Child(ObjectWithSubmenus).ui(
        action_submenus=True
    )
    map_with_submenus = flow.Child(MapWithSubmenus).ui(
        action_submenus=True,
        items_action_submenus=True,
    )
        '''
        '</pre>'
    )
    object_with_submenus = flow.Child(ObjectWithSubmenus).ui(
        action_submenus=True
    )
    map_with_submenus = flow.Child(MapWithSubmenus).ui(
        action_submenus=True,
        items_action_submenus=True,
    )



import six
from qtpy import QtWidgets, QtGui, QtCore

from ...view import ViewMixin



class DialogView(ViewMixin, QtWidgets.QDialog):

    def __init__(self, session, view_id, parent_widget):
        ViewMixin.__init__(self, session, view_id)
        QtWidgets.QDialog.__init__(self, parent_widget)

        # This is needed for layout state
        # Multi instance view types must use another policy
        self.setObjectName(self.view_type_name())


        if view_id is not None:
            self.setWindowTitle(root_oid)

        self.setProperty("dialog", True)

        self.set_on_view_title_change(self._on_view_title_change)

        self.finished.connect(self._on_dialog_finished)

    def _fix_raise_fail_on_windows(self):
        self.raise_()

    def ensure_visible(self):
        # There is a known bug on window where calling raise() works only
        # if the view is already parented+visible.
        # Known workaround is to delay the raise().
        QtCore.QTimer.singleShot(0, self._fix_raise_fail_on_windows)

    def moveToCenter(self):
        # Access to MainWindow
        app = QtWidgets.QApplication.instance()
        for widget in app.topLevelWidgets():
            if isinstance(widget, QtWidgets.QMainWindow):
                window = widget.frameGeometry()
        
        # Move dialog
        fg = self.frameGeometry()
        fg.moveCenter(window.center())
        self.move(fg.topLeft())

    def _on_dialog_finished(self):
        self.forget_dialog_view()
        self.deleteLater()

    def _on_view_title_change(self):
        self.setWindowTitle(self.view_title())

    def forget_dialog_view(self):
        self.session.forget_view(self)

    def delete_view(self):
        self.session.forget_view(self)
        self.deleteLater()

    def get_view_state(self):
        '''
        Overridden to return None, which will prevent
        the view from being restored by saved states/layouts.
        '''
        return None

class ToolBarView(ViewMixin, QtWidgets.QToolBar):

    def __init__(self, session, view_id=None, hidden=False, area=None):
        try:
            main_window = session.main_window_manager.main_window
        except AttributeError:
            raise TypeError(
                'The "%s" view cannont be used in a session without a main_window' % (
                    self.__class__.__name__
                )
            )
        ViewMixin.__init__(self, session, view_id)
        QtWidgets.QToolBar.__init__(self, main_window)

        self.hidden = hidden
        self.area = area
        self._main_window_manager = session.main_window_manager
        
        # Restore correct ToolBarArea enum value from string
        self.areaEnum = dict(
            LeftToolBarArea=QtCore.Qt.LeftToolBarArea,
            RightToolBarArea=QtCore.Qt.RightToolBarArea,
            TopToolBarArea=QtCore.Qt.TopToolBarArea,
            BottomToolBarArea=QtCore.Qt.BottomToolBarArea,
        )

        if isinstance(self.area, str):
            self.area = self.getAreaFromEnum()

        # This is needed for layout state
        # Multi instance view types must use another policy
        self.setObjectName(self.view_type_name())

        self._main_window_manager.add_toolbar_view(self, hidden=hidden, area=self.area)

        self.visibilityChanged.connect(self.toolbar_visibility_changed)

        self.set_on_view_title_change(self._on_view_title_change)
        self.set_view_title(self.view_type_name())

    def _fix_raise_fail_on_windows(self):
        self.raise_()

    def ensure_visible(self):
        # There is a known bug on window where calling raise() works only
        # if the view is already parented+visible.
        # Known workaround is to delay the raise().
        QtCore.QTimer.singleShot(0, self._fix_raise_fail_on_windows)

    def isHidden(self):
        return self.hidden

    def isVisible(self):
        return bool(self.hidden is False)

    def areaPosition(self):
        return self.area

    def getAreaFromEnum(self):
        return self.areaEnum[self.area] if self.area in self.areaEnum else None

    def toolbar_visibility_changed(self, visible):
        self.hidden = not visible
        self.area = self._main_window_manager.main_window.toolBarArea(self)

    def _on_view_title_change(self):
        self.setWindowTitle(self.view_title())

    def delete_view(self):
        self.session.forget_view(self)
        self.session.main_window_manager.main_window.removeToolBar(self)
        self.deleteLater()


class DockedView(ViewMixin, QtWidgets.QWidget):

    def __init__(self, session, view_id=None, hidden=False, area=None):
        try:
            parent = session.main_window_manager.main_window
        except AttributeError:
            raise TypeError(
                'The "%s" view cannont be used in a session without a main_window'%(
                    self.__class__.__name__
                )
            )
        ViewMixin.__init__(self, session, view_id)
        QtWidgets.QWidget.__init__(self, None)

        self.hidden = hidden
        self.area = area
        self._main_window_manager = session.main_window_manager
        
        # Restore correct DockWidgetArea enum value from string
        self.areaEnum = dict(
            LeftDockWidgetArea=QtCore.Qt.LeftDockWidgetArea,
            RightDockWidgetArea=QtCore.Qt.RightDockWidgetArea,
            TopDockWidgetArea=QtCore.Qt.TopDockWidgetArea,
            BottomDockWidgetArea=QtCore.Qt.BottomDockWidgetArea,
        )

        if isinstance(self.area, str):
            self.area = self.getAreaFromEnum()

        # Menu
        self.menubar = QtWidgets.QMenuBar(self)
        self.view_menu = QtWidgets.QMenu(self.view_title())
        self.menubar.addMenu(self.view_menu)

        # Tools
        self._header_tools = {}
        self._header_tools_layout = QtWidgets.QHBoxLayout()

        content_widget = QtWidgets.QWidget(self)

        lo = QtWidgets.QVBoxLayout()
        lo.setContentsMargins(0, 0, 0, 0)
        lo.setSpacing(0)
        self.setLayout(lo)

        hlo = QtWidgets.QHBoxLayout()
        hlo.setContentsMargins(0, 0, 0, 0)
        hlo.addWidget(self.menubar)
        header_widgets_layout = QtWidgets.QHBoxLayout()
        hlo.addStretch()
        hlo.addLayout(header_widgets_layout, 100)
        hlo.addLayout(self._header_tools_layout)
        lo.addLayout(hlo)
        top_layout = QtWidgets.QHBoxLayout()
        lo.addLayout(top_layout)
        lo.addWidget(content_widget, 100)
        self._build(
            self, top_layout, content_widget,
            self, header_widgets_layout
        )
        self._update_menus()

        dock = self._main_window_manager.create_docked_view_dock(self, hidden=hidden, area=self.area)

        # This is needed for layout state
        # Multi instance view types must use another policy
        dock.setObjectName(self.view_id())

    def _create_scrolled_area(self, parent):
        scroll_area = QtWidgets.QScrollArea(parent)
        scroll_area.setWidgetResizable(True)

        content_widget = QtWidgets.QWidget(parent)
        scroll_area.setWidget(content_widget)

        return scroll_area, content_widget

    def _build(
            self,
            top_parent, top_layout,
            main_parent,
            header_parent, header_layout
    ):
        '''
        Subclasses must override this to build their widgets.
        '''
        pass

    def duplicate_view(self, *extra_args, **extra_kwargs):
        dock = self.dock_widget()

        area = dock.dock_area()
        area = {
            QtCore.Qt.LeftDockWidgetArea: QtCore.Qt.RightDockWidgetArea,
            QtCore.Qt.RightDockWidgetArea: QtCore.Qt.BottomDockWidgetArea,
            QtCore.Qt.BottomDockWidgetArea: QtCore.Qt.LeftDockWidgetArea,
            QtCore.Qt.TopDockWidgetArea: QtCore.Qt.LeftDockWidgetArea,
        }[area]

        return self.session.add_view(
            self.view_type_name(),
            area=area,
            *extra_args, **extra_kwargs
        )

    def _fix_raise_fail_on_windows(self):
        self.dock_widget().raise_()

    def ensure_visible(self):
        # There is a known bug on window where calling raise() works only
        # if the view is already parented+visible.
        # Known workaround is to delay the raise().
        QtCore.QTimer.singleShot(0, self._fix_raise_fail_on_windows)
    
    def isHidden(self):
        return self.hidden
    
    def isVisible(self):
        return bool(self.hidden is False)
    
    def areaPosition(self):
        return self.area

    def getAreaFromEnum(self):
        return self.areaEnum[self.area] if self.area in self.areaEnum else None

    # def set_on_view_title_change(self, f):
    #     self._on_view_title_change = f

    # def set_view_title(self, title):
    #     self._view_title = title
    #     if self._on_view_title_change is not None:
    #         self._on_view_title_change()

    # def view_title(self):
    #     return self._view_title

    def add_header_tool(self, tool_id, label, tooltip, f):
        def pyside_and_pyqt_handler(checked=None, f=f):
            f()
        b = QtWidgets.QToolButton(self)
        b.setProperty('no_border', True)
        self._header_tools[tool_id] = b
        b.setText(label)
        b.clicked.connect(pyside_and_pyqt_handler)
        b.setToolTip(tooltip)
        self._header_tools_layout.addWidget(b)
        return b

    def remove_header_tool(self, tool_id):
        self._header_tools[tool_id].deleteLater()

    def _update_menus(self):
        pass

    def dock_widget(self):
        return self.parentWidget()

    def delete_on_close(self):
        '''
        If returns True, the view will be deleted when its dock is closed.

        Default is to return True unless this is the last instance of this view type.
        '''
        return self.session.view_type_count(self.view_type_name()) > 1

    def about_to_delete(self):
        '''
        Called if delete_on_close() returned True and the view is about to be deleted.
        Subclasses may override this to do some cleanup or cancel some pending commands but
        MUST call the base implementation
        '''
        self.session.forget_view(self)

    def on_hide(self):
        '''
        Called when the view's dock is hidden.
        Beware that consecutive hide/show happens when floating a view...
        '''
        pass

    def on_show(self):
        '''
        Called when the view's dock is shown.
        Beware that consecutive hide/show happens when floating a view...
        '''
        pass

    def delete_view(self):
        self.about_to_delete()
        dock = self.dock_widget()
        self.session.main_window_manager._docks.remove(dock) # UGLY protected access !
        self.session.main_window_manager.main_window.removeDockWidget(dock)
        dock.deleteLater()

    # def receive_event(self, event, data):
    #     raise NotImplementedError()













# ##### OLD #####
# class ViewType(QWidgetType):

#     def __new__(cls, class_name, bases, class_dict):
#         class_dict['_VIEWS'] = []
#         return super(ViewType, cls).__new__(cls, class_name, bases, class_dict)


# # six.with_metaclass does not play well with PySide's shiboken.
# if six.PY2:
#     class _BaseView(QtWidgets.QWidget):
#         __metaclass__ = ViewType

# else:
#     class _BaseView(six.with_metaclass(ViewType, QtWidgets.QWidget)):
#         pass


# class View(_BaseView):

#     _VIEWS = []  # will be reset to [] for all view types by the metaclass

#     @classmethod
#     def view_type_name(cls):
#         return cls.__name__

#     def __init__(self, session, main_window_manager, parent):
#         super(View, self).__init__(parent)

#         self.__class__._VIEWS.append(self)

#         self.session = session
#         self._main_window_manager = main_window_manager

#         self._view_title = self.view_type_name()
#         self._on_view_title_change = None

#         # Menu
#         self.menubar = QtWidgets.QMenuBar(self)
#         self.view_menu = self.menubar.addMenu(self.view_title())

#         # Tools
#         self._header_tools = {}
#         self._header_tools_layout = QtWidgets.QHBoxLayout()

#         content_widget = QtWidgets.QWidget(self)

#         lo = QtWidgets.QVBoxLayout()
#         lo.setContentsMargins(0, 0, 0, 0)
#         lo.setSpacing(0)
#         self.setLayout(lo)

#         hlo = QtWidgets.QHBoxLayout()
#         hlo.setContentsMargins(0, 0, 0, 0)
#         hlo.addWidget(self.menubar)
#         header_widgets_layout = QtWidgets.QHBoxLayout()
#         hlo.addStretch()
#         hlo.addLayout(header_widgets_layout, 100)
#         hlo.addLayout(self._header_tools_layout)
#         lo.addLayout(hlo)
#         top_layout = QtWidgets.QHBoxLayout()
#         lo.addLayout(top_layout)
#         lo.addWidget(content_widget, 100)
#         self._build(
#             self, top_layout, content_widget,
#             self, header_widgets_layout
#         )

#         self._update_menus()

#     def _create_scrolled_area(self, parent):
#         scroll_area = QtWidgets.QScrollArea(parent)
#         scroll_area.setWidgetResizable(True)

#         content_widget = QtWidgets.QWidget(parent)
#         scroll_area.setWidget(content_widget)

#         return scroll_area, content_widget

#     def _build(
#             self,
#             top_parent, top_layout,
#             main_parent,
#             header_parent, header_layout
#     ):
#         '''
#         Subclasses must override this to build their widgets.
#         '''
#         pass

#     def duplicate_view(self, *extra_args, **extra_kwargs):
#         dock = self.dock_widget()

#         area = dock.dock_area()
#         area = {
#             QtCore.Qt.LeftDockWidgetArea: QtCore.Qt.RightDockWidgetArea,
#             QtCore.Qt.RightDockWidgetArea: QtCore.Qt.BottomDockWidgetArea,
#             QtCore.Qt.BottomDockWidgetArea: QtCore.Qt.LeftDockWidgetArea,
#             QtCore.Qt.TopDockWidgetArea: QtCore.Qt.LeftDockWidgetArea,
#         }[area]

#         return self._main_window_manager.add_view(
#             self.view_type_name(),
#             area=area,
#             *extra_args, **extra_kwargs
#         )

#     def set_on_view_title_change(self, f):
#         self._on_view_title_change = f

#     def set_view_title(self, title):
#         self._view_title = title
#         if self._on_view_title_change is not None:
#             self._on_view_title_change()

#     def view_title(self):
#         return self._view_title

#     def add_header_tool(self, tool_id, label, tooltip, f):
#         def pyside_and_pyqt_handler(checked=None, f=f):
#             f()
#         b = QtWidgets.QToolButton(self)
#         b.setProperty('no_border', True)
#         self._header_tools[tool_id] = b
#         b.setText(label)
#         b.clicked.connect(pyside_and_pyqt_handler)
#         b.setToolTip(tooltip)
#         self._header_tools_layout.addWidget(b)
#         return b

#     def remove_header_tool(self, tool_id):
#         self._header_tools[tool_id].deleteLater()

#     def _update_menus(self):
#         pass

#     def dock_widget(self):
#         return self.parentWidget()

#     def delete_on_close(self):
#         '''
#         If returns True, the view will be deleted when its dock is closed.

#         Default is to return True unless this is the last instance of this view type.
#         '''
#         try:
#             self.__class__._VIEWS[1]
#         except IndexError:
#             return False
#         return True

#     def about_to_delete(self):
#         '''
#         Called if delete_on_close() returned True and the view is about to be deleted.
#         Subclasses may override this to do some cleanup or cancel some pending commands but
#         MUST call the base implementation
#         '''
#         self.__class__._VIEWS.remove(self)

#     def on_hide(self):
#         '''
#         Called when the view's dock is hidden.
#         Beware that consecutive hide/show happens when floating a view...
#         '''
#         pass

#     def on_show(self):
#         '''
#         Called when the view's dock is shown.
#         Beware that consecutive hide/show happens when floating a view...
#         '''
#         pass

#     def receive_event(self, event, data):
#         raise NotImplementedError()





import sys

from ...session import KabaretSession

from qtpy import QtCore, QtWidgets
from .widgets.main_window import MainWindowManager
from .widgets.flow.flow_view import FlowView
from .widgets.flow import FlowViewPlugin

class KabaretStandaloneGUISession(KabaretSession):

    def __init__(self,
        session_name='Standalone', tick_every_ms=10, debug=False,
        layout_mgr=False, layout_autosave=False, layout_savepath=None
    ):
        super(KabaretStandaloneGUISession, self).__init__(session_name, debug)
        self.main_window_manager = None
        self.tick_every_ms = tick_every_ms
        self._current_view = None

        # Layout management
        self.layout_manager = layout_mgr
        self.layout_autosave = layout_autosave
        self.layout_savepath = layout_savepath
        
        self.layout_load = False
        self.layout_sessionpath = None

        if self.layout_manager:
            self.log_info("Enabling Layout Manager")
        
        if self.layout_autosave:
            self.log_info("Enabling Layout Autosave")

    def is_gui(self):
        return True

    def register_plugins(self, plugin_manager):
        super(KabaretStandaloneGUISession, self).register_plugins(plugin_manager)
        plugin_manager.register(FlowViewPlugin, 'kabaret.flow_view')

    def create_window_manager(self):
        return MainWindowManager.create_window(self)
        
    def start(self):
        app = QtWidgets.QApplication(sys.argv)
        app.setAttribute(QtCore.Qt.AA_DontShowIconsInMenus, False)
        QtWidgets.QApplication.setOrganizationName("Supamonks")
        QtWidgets.QApplication.setApplicationName("Kabaret")
        self.main_window_manager = self.create_window_manager()
        self.main_window_manager.install()

        self.main_window_manager.main_window.setFocusPolicy(QtCore.Qt.StrongFocus)
        QtWidgets.QApplication.instance().focusChanged.connect(self.watch_focus)

        timer = QtCore.QTimer(self.main_window_manager.main_window)
        timer.timeout.connect(self.tick)
        timer.start(self.tick_every_ms)

        app.exec_()

    def watch_focus(self, old, new):
        current_view = None
        if not new:
            return

        view_type_name = FlowView.view_type_name()
        for _, view in self._views.items():
            if view.view_type_name() != view_type_name:
                continue
            children = view.findChildren(type(new), new.objectName())
            parent = view.parent()
            if (
                (new is view)
                or (children and new in children)
                or (new is parent)
            ):
                current_view = view
                break

        if not current_view:
            return

        try:
            self._current_view = current_view
            if self._current_view:
                self.dispatch_event("focus_changed", view_id=self._current_view._view_id)

        except RuntimeError:
            # Internal C++ object already deleted :/
            self._current_view = None

    def current_view(self):
        if not self._current_view:
            return self.find_view(FlowView.view_type_name())
        return self._current_view
    
    def register_view_types(self):
        '''
        Subclasses can register view types and create defaults view here.
        Use:
            type_name = self.register_view_type(MyViewType)
        to register a view type.
        
        And optionally:
            self.add_view(type_name)
        to create a default view.
        '''
        super(KabaretStandaloneGUISession, self).register_view_types()

    def _get_layout_state(self):
        return self.main_window_manager.get_layout_state()

    def _set_layout_state(self, state):
        self.main_window_manager.set_layout_state(state)
    
    def _get_geometry_state(self):
        return self.main_window_manager.get_geometry_state()
    
    def _set_geometry_state(self, state):
        self.main_window_manager.set_geometry_state(state)

    def _on_cluster_connected(self):
        '''
        Overridden to also dispatch a 'cluster_connection' event to GUI.
        '''
        super(KabaretStandaloneGUISession, self)._on_cluster_connected()
        self.dispatch_event(
            'cluster_connection',
            cluter_name=self._cluster_actor.get_cluster_name()
        )

    # def dispatch_event(self, event_type, **data):
    #     if self.main_window_manager is not None:
    #         self.main_window_manager.receive_event(event_type, data)
    #     super(KabaretStandaloneGUISession, self).dispatch_event(event_type, **data)



from kabaret.flow import (
    values,
    Object, DynamicMap, ThumbnailInfo,
    Action, ConnectAction,
    ChoiceValueSetAction, ChoiceValueSelectAction,
    Child, Parent, Computed, Connection,
    Param, IntParam, BoolParam, HashParam,
    Label
)

from kabaret.app import resources


class IconThumbnailInfo(ThumbnailInfo):

    _icon = Parent()
    _folder = Parent(2)

    def is_resource(self):
        return True

    def get_resource(self):
        return self._folder.get_icon_ref(self._icon)


class Icon(Object):

    icon_name = Param().ui(editable=False)
    _thumbnail_info = Child(IconThumbnailInfo)

    def get_thumbnail_object(self):
        return self._thumbnail_info


class Icons(DynamicMap):

    _icon_folder = Parent()

    @classmethod
    def mapped_type(cls):
        return Icon

    def __init__(self, parent, name):
        super(Icons, self).__init__(parent, name)
        self._items = dict(
            [
                ('i_' + n.replace('.', '_').replace('-', '_'), n)
                for n in resources.list_folder(self.folder_name())
            ]
        )

    def mapped_names(self, page_num=0, page_size=None):
        return list(self._items.keys())

    def _create_child(self, name, object_type):
        '''
        Creates and return a new child object of the given type.
        '''
        child = object_type(self, name)
        icon_name = self._items[name]
        child.icon_name.set(icon_name)
        return child

    def folder_name(self):
        return self._icon_folder.folder_name.get()

    def get_icon_ref(self, icon_item):
        folder_name = self.folder_name()
        icon_name = icon_item.icon_name.get()
        return (folder_name, icon_name)

    def columns(self):
        return ['Name', 'Ref']

    def _fill_row_cells(self, row, item):
        row['Name'] = ''
        row['Ref'] = repr((self.folder_name(), item.icon_name.get()))

    def _fill_row_style(self, style, item, row):
        style['icon'] = self.get_icon_ref(item)


class IconFolder(Object):

    folder_name = Param().ui(editable=False)

    icons = Child(Icons).ui(expanded=True)


class IconFolders(DynamicMap):

    @classmethod
    def mapped_type(cls):
        return IconFolder

    def __init__(self, parent, name):
        super(IconFolders, self).__init__(parent, name)
        self._items = dict(
            [
                (n.replace('.', '_').replace('-', '_'), n)
                for n in resources.get_folder_names()
            ]
        )

    def mapped_names(self, page_num=0, page_size=None):
        return list(self._items.keys())

    def _create_child(self, name, object_type):
        '''
        Creates and return a new child object of the given type.
        '''
        child = object_type(self, name)
        folder_name = self._items[name]
        child.folder_name.set(folder_name)
        return child


class IconBrowserGroup(Object):

    doc = Label('<H2>Here you can browse icons from the resources folder.</H2>')
    folder_names = Child(IconFolders).ui(expanded=True)



import six
from qtpy import QtWidgets, QtCore

from kabaret.app import resources

from ..form_tree import FormField
from .ref_by import RefByWidget


class ObjectActionMenuManager(object):

    def __init__(self, session, action_callback, context):
        super(ObjectActionMenuManager, self).__init__()
        self.session = session
        self.action_callback = action_callback
        self.context = context

    def _update_menu(self, menu, actions, action_callback, with_submenus):
        group_to_menu = {None:menu}
        def get_menu(names):
            try:
                return group_to_menu[tuple(names)]
            except KeyError:
                if not names:
                    return menu
                parent = get_menu(names[:-1])
                submenu = parent.addMenu(names[-1])
                group_to_menu[tuple(names)] = submenu
                return submenu
        menu.clear()
        last_group = None
        current_menu = menu
        for group, label, data, enabled, icon_ref in actions:
            if last_group != group:
                if with_submenus:
                    names = group and group.split('/') or group
                    current_menu = get_menu(names)
                else:
                    menu.addSeparator()
            last_group = group
            action_oid = data['oid']
            menu_callback = lambda oid=action_oid: action_callback(oid)  # noqa
            action = current_menu.addAction(label, menu_callback)
            if not enabled:
                action.setEnabled(False)
            if icon_ref and icon_ref[1]:
                try:
                    icon = resources.get_icon(icon_ref, self)
                except resources.NotFoundError as err:
                    self.session.log_debug("WARNING: RESOURCE NOT FOUND: %r" % (err,))
                else:
                    action.setIcon(icon)

    def update_oid_menu(self, oid, menu, context=None, with_submenus=False):
        context = context or self.context
        actions_info = self.session.cmds.Flow.get_object_actions(oid, context)
        # =>  (
        #         relation.index,
        #         ui_group,
        #         relative_action_path,
        #         {
        #             'action_icon': relation_related_type.ICON,
        #             'ui': action_ui_config,
        #             'oid': action_oid,
        #         }
        #     )

        actions = []
        for index, group, path, data in actions_info:
            enabled = data.get('ui', {}).get('enabled', True)
            icon_name = data.get('action_icon') or 'action'
            if isinstance(icon_name, six.string_types):
                icon_ref = ('icons.flow', icon_name)
            else:
                icon_ref = icon_name
            label = data.get('ui', {}).get('label')
            if len(path) > 1 or label is None:
                label = ' > '.join(
                    name.strip('_').replace('_', ' ').title()
                    for name in path
                )
            actions.append((
                group,
                label,
                data,
                enabled,
                icon_ref
            ))

        self._update_menu(menu, actions, self.action_callback, with_submenus=with_submenus)

        return bool(actions)

    def update_oids_menu(self, oids, menu, context=None, with_submenus=False):
        context = context or self.context
        actions_info = self.session.cmds.Flow.get_objects_actions(oids, context)

        actions = []
        for index, group, path, data in actions_info:
            enabled = data.get('ui', {}).get('enabled', True)
            icon_name = data.get('action_icon') or 'action'
            if isinstance(icon_name, six.string_types):
                icon_ref = ('icons.flow', icon_name)
            else:
                icon_ref = icon_name
            label = data.get('ui', {}).get('label')
            if len(path) > 1 or label is None:
                label = ' > '.join(
                    name.strip('_').replace('_', ' ').title()
                    for name in path
                )
            actions.append((
                group, label, data, enabled, icon_ref
            ))

        self._update_menu(
            menu, actions, self._multiple_action_callback, 
            with_submenus=with_submenus
        )
        return bool(actions)

    def _multiple_action_callback(self, oids):
        for oid in oids:
            self.action_callback(oid)


class ObjectActionsMenu(QtWidgets.QWidget):

    def __init__(self, session, parent, action_callback, context):
        super(ObjectActionsMenu, self).__init__(parent)
        self._menu_manager = ObjectActionMenuManager(
            session, action_callback, context
        )
        self.session = session

        self._with_submenus = False

        self.setLayout(QtWidgets.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        tb = QtWidgets.QToolButton(self)
        tb.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        tb.setArrowType(QtCore.Qt.NoArrow)
        tb.setProperty('hide_arrow', True)
        tb.setProperty('no_border', True)
        tb.setPopupMode(tb.ToolButtonPopupMode.InstantPopup)
        try:
            icon = resources.get_icon(('icons.gui', 'submenu'), self)
        except resources.NotFoundError as err:
            self.session.log_debug("WARNING: RESOURCE NOT FOUND: %r" % (err,))
        else:
            tb.setIcon(icon)

        self._menu = QtWidgets.QMenu(tb)
        tb.setMenu(self._menu)
        self.layout().addWidget(tb)

        self.hide()
        # self.set_actions(None)

    def set_with_submenus(self, b):
        self._with_submenus = bool(b)

    def load_actions(self, oid):
        got_actions = self._menu_manager.update_oid_menu(
            oid, self._menu, with_submenus=self._with_submenus
        )
        if not got_actions:
            self.hide()
        else:
            self.show()
        return


class ObjectSummary(QtWidgets.QWidget):

    def __init__(self, session, parent, icon_provider=None):
        super(ObjectSummary, self).__init__()
        self.session = session
        self.icon_provider = icon_provider

        lo = QtWidgets.QHBoxLayout()
        lo.setContentsMargins(0, 0, 0, 0)
        self.setLayout(lo)

        self._tb = QtWidgets.QToolButton(self)
        self._tb.hide()
        self._lb = QtWidgets.QLabel(self)

        lo.addWidget(self._tb)
        lo.addWidget(self._lb)

    def load_summary(self, oid):
        summary = self.session.cmds.Flow.get_summary(oid)
        self._lb.setText(summary)
        if self.icon_provider is not None:
            icon = self.icon_provider(summary)
            self.set_icon(icon)

    def set_icon(self, icon):
        if icon is None:
            self._tb.hide()
            return
        self._tb.setIcon(icon)
        self._tb.show()


class FlowField(FormField):

    def __init__(self, parent, session, oid, ui_config=None):
        self.session = session
        self.oid = oid
        self.ui_config = ui_config or {}
        self._label = self.ui_config.get('label')
        if self._label is None:
            self._label = oid.rsplit('/', 1)[-1].replace('_', ' ').title()

        super(FlowField, self).__init__(parent)

    def auto_expand(self):
        if self.ui_config.get("expanded") and self._pref_id() not in self.__class__._LAST_EXPANDED_STATES:
            self.setExpanded(True)
        else:
            super(FormField, self).auto_expand()
            
    def show_may_have_children(self):
        if self.ui_config.get('expandable', True):
            super(FlowField, self).show_may_have_children()

    def _get_icon(self, icon_name, resource_folder='icons.flow'):
        if isinstance(icon_name, six.string_types):
            icon_ref = (resource_folder, icon_name)
        else:
            icon_ref = icon_name
        try:
            icon = resources.get_icon(icon_ref, self.treeWidget())
        except resources.NotFoundError as err:
            self.session.log_debug("WARNING: RESOURCE NOT FOUND: %r" % (err,))
            return None
        return icon

    def _get_config_icon(self):
        icon_name = self.ui_config.get('icon', 'object')
        if not icon_name:
            return None
        return self._get_icon(icon_name)

    def _get_icon_for(self, text, resource_folder='icons.status', icon_name=None):
        if icon_name is not None:
            return self._get_icon(icon_name, resource_folder)

        if text is None:
            return None
        first_word = text.split(' ', 1)[0]
        return self._get_icon(first_word.upper(), resource_folder)

    def activated(self, col):
        if QtWidgets.QApplication.keyboardModifiers() & QtCore.Qt.ControlModifier:
            self.treeWidget().open(self.oid)
            return True
        return None

    def build_children(self):
        tree = self.treeWidget()
        tree.build_children(self)

    def on_touch_event(self, oid):
        # Since an object can appear in several pages using the `Relative`
        # relation, we cant stop propagating once the event is handlded
        # so the return value is not used here:
        if self._inspect_touch_event(oid) and False:
            return True
        return self._dispatch_touch_event(oid)

    def _inspect_touch_event(self, oid):
        if oid == self.oid:
            # print('  My OTE', self.oid)
            self.update_content()
            return True
        return False

    def _dispatch_touch_event(self, oid):
        # We use to prevent event propagation when the event
        # did not point to an oid under this field, so that
        # propagation stop quickly.
        # But using DynamicMap to display items that are not
        # children of the map, or having Ref showing children
        # that are not children of the Ref (but children of the
        # target), is not compatible with this optimisation.
        # As this optimisation is not essential, we now just
        # dispatch everywhere:
        if 1 or oid.startswith(self.oid + '/'):
            # print('  Child OTE', self.oid)
            for i in range(self.childCount()):
                c = self.child(i)
                try:
                    c.on_touch_event
                except AttributeError:
                    # This occurs over "_PRELOAD_CHILD_" :/ (who said ugly? :p)
                    pass
                else:
                    handled = c.on_touch_event(oid)
                    # Since an object can appear in several pages using the `Relative`
                    # relation, we cant stop propagating once the event is handlded
                    # so this is disabled:
                    if 0 and handled:
                        return True
        return False

    def update_content(self):
        raise NotImplementedError(self.__class__.__name__)


class ReferencesField(FormField):

    _DEFAULT_EXPANDED_STATE = True

    def __init__(self, parent, oids, leaf_callback, group_callback):
        self.ref_by_widget = None
        self.oids = oids
        self._leaf_callback = leaf_callback
        self._group_callback = group_callback
        super(ReferencesField, self).__init__(parent)

    def activated(self, col):
        if not self.isExpanded():
            self.setExpanded(True)
        else:
            self.setExpanded(False)

    def build_children(self):
        return

    def build(self):
        tree = self.treeWidget()

        self.setFlags(QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable)

        self.setIcon(0, resources.get_icon(('icons.flow', 'ref')))
        self.setText(0, "References")
        self.ref_by_widget = RefByWidget(tree, self.oids,
                                         self._leaf_callback, self._group_callback,
                                         self.on_tree_hide)
        self.setItemWidget(1, self.ref_by_widget)

    def on_tree_hide(self, hidden):
        if self.ref_by_widget:
            self.update_height(self.ref_by_widget)

    def on_touch_event(self, oid):
        self.ref_by_widget.refresh()
        return True



import os
import re
from .interface import Editor_Interface

from six import text_type as unicode

from qtpy import QtWidgets, QtCore, QtGui
from kabaret.app import resources


class FileDialog(QtWidgets.QFileDialog):

    def __init__(self, options):
        QtWidgets.QFileDialog.__init__(self)

        # Select multiple files
        if options.get('files_only') is True:
            self.setFileMode(self.FileMode.ExistingFiles)
            # Select only one file
            if options.get('max_count') == 1:
                self.setFileMode(self.FileMode.ExistingFile)
        # Select only one directory
        elif options.get('dirs_only') is True and options.get('max_count') == 1:
            self.setFileMode(self.FileMode.Directory)
        # Select multiple files and directories or multiple directories
        else:
            self.setFileMode(self.FileMode.Directory)
            self.setOption(self.Option.DontUseNativeDialog, True)
            self.setViewMode(self.ViewMode.Detail)

            # Allow multi selection in views
            self.list = self.findChild(QtWidgets.QListView)
            self.list.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection)

            self.tree = self.findChild(QtWidgets.QTreeView)
            self.tree.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection)

            # Redefine clicked signal for main button
            # To fix a issue when the first item is a file
            btns = self.findChildren(QtWidgets.QPushButton)
            self.chooseBtn = [x for x in btns if 'choose' in str(x.text()).lower()][0]
            self.chooseBtn.installEventFilter(self)
            self.chooseBtn.clicked.disconnect()
            self.chooseBtn.clicked.connect(self.openClicked)
            
            if options.get('dirs_only') is True:
                self.setOption(self.Option.ShowDirsOnly, True)
        
        # Select the current directory path
        if options.get('start_path'):
            self.setDirectory(options.get('start_path'))
        
        # Sets the filters
        if options.get('filters'):
            self.setNameFilters(options.get('filters'))

    def eventFilter(self, source, event):
        # Keep enabled main button
        if isinstance(source, QtWidgets.QPushButton) and event.type() is QtCore.QEvent.EnabledChange:
            if source.isEnabled() is False:
                source.setEnabled(True)
                return True
        
        return False

    def openClicked(self):
        # Fetch selected files
        indexs = self.tree.selectionModel().selectedIndexes()
        files = []
        for i in indexs:
            if i.column() == 0:
                files.append(os.path.normpath(
                    os.path.join(self.directory().absolutePath(), str(i.data()))
                ))
        
        # Set dialog result to Accepted
        if files:
            self.setResult(1)

        self.selectedFiles = files
        self.hide()


class PathValueEditor(QtWidgets.QLineEdit, Editor_Interface):
    '''
    This editor lets you enter path values using the file explorer.
    By default, the user can select any files.

    Editor Type Names:
        path

    Options:
        start_path:      str - Open file explorer from a specific path
        filters:         list of str (ex: ['Text files (*.txt)', 'Any files (*)']) - Displays files that match the patterns
        max_count:       int - Limit the maximum number of paths
        dirs_only:       bool - Can only select directories
        files_only:      bool - Can only select files
        path_format:     str ('win32' or 'linux') - Force path format to a specific operating system
    '''

    @classmethod
    def can_edit(cls, editor_type_name):
        '''
        Must be implemented to return True if the given editor_type_name
        matches this editor.
        '''
        return editor_type_name in ('path',)

    def __init__(self, parent, options):
        QtWidgets.QLineEdit.__init__(self, parent)
        Editor_Interface.__init__(self, parent)

        self.options = options
        
        # Default regex
        self.regexp = '^(?:[a-zA-Z]:[\\\\\/]{{1,2}}|\/{{1,2}}|\\{{1,2}})(?:[^<>:"/\\\\|?*]+[\\\\\/]{{1,2}})*(?:[^<>:"/\\\\|?*]+{filters})(?<=\S)$'

        self.setLayout(QtWidgets.QHBoxLayout())
        self.layout().setContentsMargins(5,5,5,5)

        self._explorer_button = QtWidgets.QToolButton(self)
        self._explorer_button.setIcon(QtGui.QIcon(resources.get_icon(('icons.gui', 'open-folder'))))
        self._explorer_button.setFixedSize(QtCore.QSize(20,20))
        self._explorer_button.setIconSize(QtCore.QSize(12,12))
        self._explorer_button.setCursor(QtCore.Qt.ArrowCursor)
        self._explorer_button.clicked.connect(self._open_explorer)
       
        self.layout().addStretch()
        self.layout().addWidget(self._explorer_button)

        self.apply_options(options)

    def set_editable(self, b):
        '''
        Must be implemented to prevent editing if b is False.
        Visual cue show also be given to the user.
        '''
        self.setReadOnly(not b)
        if b:
            self.textEdited.connect(self._on_edited)
            self.returnPressed.connect(self.apply)
        else:
            self._explorer_button.hide()

    def apply_options(self, options):
        '''
        Must be implemented to configure the editor as
        described by the options dict.
        '''
        # Check value types of options
        if options.get("start_path") and type(options.get("start_path")) is not str:
            raise Exception('start_path option should be setted with a str value.')

        if options.get("max_count") and type(options.get("max_count")) is not int:
            raise Exception('max_count option should be setted with a int value.')

        if options.get("dirs_only") and type(options.get("dirs_only")) is not bool:
            raise Exception('dirs_only option should be setted with a bool value.')

        if options.get("files_only") and type(options.get("files_only")) is not bool:
            raise Exception('files_only option should be setted with a bool value.')

        if options.get("dirs_only") and options.get("files_only"):
            raise Exception('dirs_only and files_only options can\'t be used at the same time.')
        
        # Use specific regex if path_format option
        if options.get("path_format") is not None:
            if options.get("path_format") == "win32":
                self.regexp = '^(?:[a-zA-Z]:[\\\\\/]{{1,2}}|\\{{2}})(?:[^<>:"/\\\\|?*]+[\\\\\/]{{1,2}})*(?:[^<>:"/\\\\|?*]+{filters})(?<=\S)$'
            elif options.get("path_format") == "linux":
                self.regexp = '^\/(?:[^/]+\/)*[^/]+{filters}(?<=\S)$'
            else:
                raise Exception('path_format option must be "win32" or "linux".')
        
        # Update regex if filters (file extensions)
        if options.get('filters'):
            filters = re.findall('(?<=\*.)\w+|\(\*\)', str(options.get('filters')))
            if '(*)' not in filters:
                self.regexp = self.regexp.format(
                    filters='\.(?:{exts})'.format(
                        exts='|'.join(filters)
                    )
                )
                return
        
        self.regexp = self.regexp.format(filters='')

    def update(self):
        '''
        Must be implemnented to show the value returned by self.fetch_value()
        Your code should call self._on_updated() at the end.
        '''
        # Clear ToolTip if there was an error
        self.setToolTip('')

        if self.text():
            # For multiple paths
            if self.text().startswith("['") and self.text().endswith("']"):
                paths = eval(self.text())

                # Check maximum count
                if self.options.get('max_count'):
                    if len(paths) > self.options.get('max_count'):
                        word = f'path{"s" if self.options.get("max_count") > 1 else ""}'
                        return self._show_error(f'A maximum of {self.options.get("max_count")} {word} can be defined')
                
                # Validate path format
                for i, path in enumerate(paths):
                    if re.search(self.regexp, path) is None:
                        # Specify which format if path_format option
                        if self.options.get('path_format'):
                            return self._show_error(f'Path {i+1} must be in {self.options.get("path_format")} format')
                        else:
                            return self._show_error(f'Path {i+1} is not valid')
            # For single path
            else:
                # Validate path format
                if re.search(self.regexp, self.text()) is None:
                    # Specify which format if path_format option
                    if self.options.get('path_format'):
                        return self._show_error(f'Path must be in {self.options.get("path_format")} format')
                    else:
                        return self._show_error('Path is not valid')
        
        self.setText(unicode(self.fetch_value()))
        self._on_updated()

    def get_edited_value(self):
        '''
        Must be implemented to return the value currently displayed.
        '''
        value = self.text()
        if not value.isalpha():     # to avoid python keywords
            try:
                value = eval(value)
            except Exception:
                pass
        return value

    def _open_explorer(self):
        fd = FileDialog(self.options)

        if fd.exec_():
            paths = fd.selectedFiles() if callable(fd.selectedFiles) else fd.selectedFiles
            self.setText(paths[0] if len(paths) == 1 else str(paths))

            self._on_edited()

            if self.options.get('max_count'):
                if len(paths) > self.options.get('max_count'):
                    word = f'path{"s" if self.options.get("max_count") > 1 else ""}'
                    self._show_error(f'A maximum of {self.options.get("max_count")} {word} can be defined')
    
    def _show_edited(self):
        '''
        Must be implemented to show that the displayed value
        needs to be applied.
        '''
        self.setProperty('edited', True)
        self.setProperty('applying', False)
        self.setProperty('error', False)
        self.style().polish(self)

    def _show_applied(self):
        '''
        Must be implemented to show that the displayed value
        as been saved.
        In a clean scenario, applying edits will trigger an update()
        and this state should disapear.
        If you are using the Editor without this kind of round trip,
        you can call update here.
        '''
        self.setProperty('applying', True)

    def _show_clean(self):
        '''
        Must be implemented to show that the displayed value is 
        up to date.
        '''
        self.setProperty('edited', False)
        self.setProperty('applying', False)
        self.setProperty('error', False)
        self.style().polish(self)

    def _show_error(self, error_message):
        '''
        Must be implemented to show that the given error occured.
        '''
        self.setProperty('error', True)
        self.style().polish(self)
        self.setToolTip('!!!\nERROR: %s' % (error_message,))

    def dragEnterEvent(self, event):
        data = event.mimeData()
        urls = data.urls()
        if urls[0].scheme() == 'file':
            event.acceptProposedAction()

    def dragMoveEvent(self, event):
        data = event.mimeData()
        urls = data.urls()
        if urls[0].scheme() == 'file':
            event.acceptProposedAction()
            cursor_pos = self.cursorPositionAt(event.pos())
            self.setCursorPosition(cursor_pos)

    def dropEvent(self, event):
        data = event.mimeData()
        urls = data.urls()
        if urls[0].scheme() == 'file':
            filepath = str(urls[0].toLocalFile()) if len(urls) == 1 else str([url.toLocalFile() for url in urls])
            self.insert(filepath)
            self._on_edited()




import time

from kabaret.flow import (
    values,
    Object, Map, Action, ConnectAction,
    ChoiceValueSetAction, ChoiceValueSelectAction,
    Child, Parent, Computed, Connection,
    Param, IntParam, BoolParam, HashParam,
    Label,
)


class MyChoices(values.ChoiceValue):

    CHOICES = ['WIP', 'RVW', 'RTK', 'DONE']


class MyMultiChoices(values.ChoiceValue):

    CHOICES = ['WIP', 'RVW', 'RTK', 'DONE']
    STRICT_CHOICES = False


class EditorsGroup(Object):

    doc = Label(
        '<HR><H4>'
        'Every relation can have its GUI representation configured:'
        '</H4>'
        '<pre>\n'
        '\tclass MyObject(Object):\n'
        '\t  my_child = Param(MyChild).ui(hidden=True, label="The Child", group="Group Name")'
        '</pre>\n'
        'Base options are:'
        '<ul>'
        '   <li>editable (bool)\t\t- set the editor writable or read-only</li>'
        '   <li>tooltip (str)\t\t- define the tooltip content</li>'
        '   <li>label (str)\t\t- define the editor label</li>'
        '   <li>group (str)\t\t- define the editors group where the editor will be</li>'
        '   <li>hidden (bool)\t\t- if True, the editor will be hidden unless the "show hidden value" is activated</li>'
        '</ul>'
        '<H4>'
        'Additionnaly to the group, label and visibility, you can control<br>'
        'which "Editor" is used when the related Object is a Value.<br>'
        'Here are examples of the default available editors.'
        '</H4>'
    )

    text_doc = Label(
        '<HR><H4>'
        'By <u>default</u> it displays a text field'
        '</H4>'
        'Options are:'
        '<ul>'
        '   <li>placeholder (str)</li>'
        '</ul>'
    )
    text = Param("").ui(placeholder="Enter your value")

    bool_doc = Label(
        '<HR><H4>'
        'The <u>bool</u> displays a check box'
        '</H4>'
        'Options are:'
        '<ul>'
        '   <li>true_text (str)</li>'
        '   <li>false_text (str)</li>'
        '</ul>'
        '<H3>Example:</H3>'
        '<pre>\n'
        '\t  bool_value = Param(False).ui(editor="bool", true_text="This is True", false_text="This is False")'
        '</pre>\n'
    )
    bool_value = BoolParam(False)  # default editor is 'bool' for BoolValue
    bool_value_with_text = Param(False).ui(
        editor='bool',
        true_text='This is True', false_text='This is False'
    )  # you can use the 'bool' editor even if not BoolValue

    choice_doc = Label(
        '<HR><H4>'
        'The <u>choice</u> displays a choice menu'
        '</H4>'
        'Options are:'
        '<ul>'
        '   <li>sorted (bool)</li>'
        '   <li>choice_icons:   dict mapping each choice to an icon_ref<br>'
        '       example: choice_icons={"Work in Progress":("icons.status", "WIP")}'
        '</ul>'
        '<H3>Example:</H3>'
        '<pre>\n'
        '\tclass MyChoices(values.ChoiceValue):\n'
        '\t  CHOICES = ["WIP", "RVW", "RTK", "DONE"]\n'
        '\t(...)\n'
        '\t  # defaut editor is "choice" for ChoiceValue\n'
        '\t  choice_value = Param("WIP", MyChoices)'
        '</pre>\n'
    )
    # defaut editor id 'choice' for ChoiceValue
    choice_value = Param('WIP', MyChoices)

    choice_icon_doc = Label(
        'The choice editor tries to find a status icon matching the first word for each choice<br>'
        'You can also provide arbitrary icon for each choice with the "choice_icons" option:<br>'
        '<pre>\n'
        '\tstatus = Param("DONE", MyChoices).ui(\n'
        '\t    choice_icons={\n'
        '\t        "WIP":("icons.flow","asset"), \n'
        '\t        "RVW":("icons.flow","film"), \n'
        '\t        "RTK":("icons.flow","shot"), \n'
        '\t        "DONE":("icons.gui","kabaret_icon"), \n'
        '\t    }\n'
        '\t)\n'
        '</pre>\n'
    )
    iconed_choice_value = Param('DONE', MyChoices).ui(
        choice_icons={
            'WIP':('icons.flow', 'asset'),
            'RVW':('icons.flow', 'film'),
            'RTK':('icons.flow', 'shot'),
            'DONE':('icons.gui', 'kabaret_icon'),
        }
    )
    text_choice_doc = Label(
        '<H3>'
        'You can still use another editor if you prefer'
        '</H3>'
        '<H3>Example:</H3>'
        '<pre>\n'
        '\t  text_choice_value = Param("WIP", MyChoices).ui(editor="python")'
        '</pre>\n'
    )
    text_choice_value = Param('WIP', MyChoices).ui(editor='python')

    multichoice_doc = Label(
        '<HR><H4>'
        'The <u>multichoice</u> editor let you select more than one choice</H4>'
        'Options are:'
        '<ul>'
        '   <li>sorted (bool)</li>'
        '</ul>'
        '<H3>Example:</H3>'
        '<pre>\n'
        '\tclass MyMultiChoices(values.ChoiceValue):\n'
        '\t  CHOICES = ["WIP", "RVW", "RTK", "DONE"]\n'
        '\t  STRICT_CHOICES = False\n'
        '(...)\n'
        '\t  multichoice_value = Param(["WIP"], editor="multichoice",'
        ' tooltip="Select one or more item in this awesome list !")'
        '</pre>\n'
    )
    multichoice = Param(['WIP'], MyMultiChoices).ui(
        editor='multichoice', tooltip='Select one or more item in this awesome list !'
    )

    multichoice_default_height_doc = Label(
        'In some cases you will need to control the multichoice default height.'
        'This can be achived with the <b>default_height</b> option'
    )
    multichoice_small = Param(['WIP'], MyMultiChoices).ui(
        editor='multichoice', tooltip='Select one or more item in this awesome list !',
        default_height=10
    )
    multichoice_big = Param(['WIP'], MyMultiChoices).ui(
        editor='multichoice', tooltip='Select one or more item in this awesome list !',
        default_height=150
    )

    textarea_doc = Label(
        '<HR><H4>'
        'The <u>textarea</u> editor shows multiple lines of text, and can be resized '
        'with the middle button.<br>'
        'Once edited, you can Apply your modifications or Cancel them with the appropriate button '
        'or shortcut (Ctrl+Renter / Escape)'
        '</H4>'
        'Options are:'
        '<ul>'
        '   <li>html (bool)</li>'
        '   <li>buttons_side  top/bottom/left/right (defaults to right)</li>'
        '</ul>'
        '<H3>Example:</H3>'
        '<pre>\n'
        '\t  textarea = Param("This is a html Multiline'
        '&lt;br&gt;&lt;font color="red"&gt;NOT Editable !&lt;/font&gt;", editor="textarea",'
        '\n\t    tooltip="This text is not editable... (editable=False)"), editable=False, html=True)'
        '</pre>\n'
    )
    textarea = Param('This\nIs\nMultiline\nEditable').ui(
        editor='textarea',
        tooltip='This text is editable, hit "Apply" to save.'
    )
    read_only_textarea = Param('This is a html Multiline<br><font color="red">NOT Editable !</font>').ui(
        editor='textarea',
        editable=False,
        tooltip='This text is not editable... (editable=False)',
        html=True
    )
    textarea_buttons_side = Param(
        'This textarea buttons are on the bottom side (edit this text to show them)\n'
        'This is achieved using option "buttons_side" with a value of '
        'top/bottom/left/right'
    ).ui(
        editor='textarea', buttons_side='bottom',
        tooltip='Edit the text to show the buttons.'
    )

    password_doc = Label(
        '<HR><H4>'
        'The <u>password</u> editor shows a text field with password format (hidden characters)<br>'
        '</H4>'
        'Options are:'
        '<ul>'
        '   <li>placeholder (str)</li>'
        '</ul>'
        '<H3>Example:</H3>'
        '<pre>\n'
        '\t  password = Param("password").ui(editor="password")'
        '</pre>\n'
    )
    password = Param("password").ui(editor="password")

    date_doc = Label(
        '<HR><H4>'
        'The <u>datetime</u> editor shows a date with the given format </br>'
        'You can edit the date with a calendar </br>'
        'The value stored is in seconds since epoch'
        '</H4>'
        'Options are:'
        '<ul>'
        '   <li>format (str)</li>'
        '   <li>minimum (seconds since epoch)</li>'
        '   <li>maximum (seconds since epoch)</li>'
        '</ul>'
        '<H3>Example:</H3>'
        '<pre>\n'
        '\t  datetime = Param(time.time()).ui(editor="datetime", format="MM/yyyy HH:ss")'
        '</pre>\n'
    )
    datetime = Param(time.time()).ui(editor="datetime", format="MM/yyyy HH:ss")
    date = Param(time.time()).ui(editor="date")

    progress_doc = Label(
        '<HR><H4>'
        'The <u>percent</u> editor shows a progress bar<br>'
        '</H4>'
        'Options are:'
        '<ul>'
        '   <li>minimum (float)</li>'
        '   <li>maximum (float)</li>'
        '   <li>display_text (bool) - display also progression in text or not'
        '</ul>'
        '<H3>Example:</H3>'
        '<pre>\n'
        '\t  progress = Param(75).ui(editor="percent")'
        '</pre>\n'
    )
    progress = Param(75).ui(editor="percent")

    error_doc = Label(
        '<HR><H4>'
        'Error are handled mostly with the tooltip and the property error="true" which can be handled in stylesheet'
        '</H4>'
        '<H3>Example:</H3>'
        '<pre>\n'
        '\t  date_error = Param(object()).ui(editor="datetime")'
        '</pre>\n'
    )
    date_error = Param(object()).ui(editor="datetime")

    thumbnail_doc = Label(
        '<HR><H4>'
        'The <u>thumbnail</u> editor shows an image which can be modified by a snapshot<br>'
        'The editor will not work if no path is set'
        '</H4>'
        '<H3>Example:</H3>'
        '<pre>\n'
        '\t  thumbnail = Param(<path_to_image>).ui(editor="thumbnail")'
        '</pre>\n'
    )
    thumbnail = Param(None).ui(editor="thumbnail")

    path_doc = Label(
        '<HR><H4>'
        'The <u>path</u> editor shows a text field that can be defined with a path value or multiple paths in a list.<br>'
        'A button on the right lets you use the file explorer to find a file or directory.'
        '</H4>'
        'Options are:'
        '<ul>'
        '   <li>start_path (str) - Open file explorer from a specific path</li>'
        '   <li>filters (list of str) - Displays files that match the patterns<br>'
        '       example: filters=["Text files (*.txt)", "Any files (*)"]</li>'
        '   <li>max_count (int) - Limit the maximum number of paths</li>'
        '   <li>dirs_only (bool) - Can only select directories</li>'
        '   <li>files_only (bool) - Can only select files</li>'
        '   <li>path_format win32/linux - Force path format to a specific operating system</li>'
        '</ul>'
        'Note: Depending on the options used, the operating system\'s native file explorer will be used instead.'
        '<H3>Example:</H3>'
        '<pre>\n'
        '\tpath = Param().ui(editor="path")'
        '</pre>\n'
    )
    path = Param().ui(editor="path")

    image_doc = Label(
        '<HR><H4>'
        'The <u>image</u> editor lets you define an image to be converted to a base64 value.<br>'
        'You can click to use the file explorer or drag and drop an image file.'
        '</ul>'
        '<H3>Example:</H3>'
        '<pre>\n'
        '\timage = Param().ui(editor="image")'
        '</pre>\n'
    )
    image = Param().ui(editor="image")




import redis.commands
import six
import redis
import os

from .._actor import Actor, Cmd, Cmds



class ReadWriteRedis:
    '''
    This is a wrapper around two `redis.Redis`
    which redirect commands to one of the two connection
    if it's a read command.
    '''

    _overridden_methods = {'ping', 'set_name'}

    #TODO: probably add more commands !!!
    _read_commands = {
        # redis.commands.core.BasicKeyCommands:
        'bitcount', 'bitfield', 'bitfield_ro', 'bitop', 'bitpos', 
        'dump', 'exists', 'expiretime', 
        'get', 'getbit', 'getrange', 'keys', 
        'mget', 
        'pexpiretime', 'pttl', 'hrandfield', 'randomkey', 
        'strlen', 'substr', 
        'ttl', 'type', 
        #FIXME: do we want to get these from read replicat?:  watch, unwatch,
        'lcs',
        # redis.commands.core.HyperlogCommands
        # redis.commands.core.HashCommands
        'hexists', 'hget', 'hgetall', 'hkeys', 'hlen', 'hmget', 'hvals', 'hstrlen',
        # redis.commands.core.GeoCommands
        # redis.commands.core.ListCommands
        'lindex', 'llen', 'lrange', 
        # redis.commands.core.ScanCommands
        # redis.commands.core.SetCommands
        'scard', 'sdiff', 'sinter', 'sintercard', 'sismember', 'smembers', 'smismember', 'srandmember', 'sunion', 
        # redis.commands.core.StreamCommands
        # redis.commands.core.SortedSetCommands
        'zcard', 'zcount', 'zdiff', 'zinter', 'zintercard', 'zlexcount', 'zrandmember', 
        'zrange', 'zrevrange', 'zrangebylex', 'zrevrangebylex', 'zrangebyscore', 'zrevrangebyscore', 
        'zrank', 'zrevrank', 'zscore', 'zunion', 'zmscore', 
    }
    def __init__(self, logger, read_db: redis.Redis, db: redis.Redis):
        super().__init__()
        self._logger = logger
        self._read_db = read_db
        self._db = db

    def __getitem__(self, name):
        self._logger(f'Using read replica for `get` command.')
        return self._read_db.__getitem__(name)
    
    def __setitem__(self, name, value):
        self._logger(f'Using master DB for `set` command.')
        self._db[name] = value

    def __delitem__(self, name):
        self._logger(f'Using master DB for `del` command.')
        return self._db.__delitem__(name)
    
    def __getattribute__(self, name: str):
        if name.startswith('_') or name in self._overridden_methods:
            return super().__getattribute__(name)
        
        if name in self._read_commands:
            self._logger(f'Using read replica for {name!r} command.')
            return getattr(self._read_db, name)
        self._logger(f'Using master DB for {name!r} command.')
        return getattr(self._db, name)
    
    def ping(self, **kwargs):
        self._logger('Pinging both db.')
        self._read_db.ping(**kwargs)
        self._db.ping(**kwargs)

    def client_setname(self, name, **kwargs):
        self._logger('Setting name on both db.')
        self._read_db.client_setname(name, **kwargs)
        self._db.client_setname(name, **kwargs)

class ClusterCmds(Cmds):

    pass


@ClusterCmds.cmd
class Connect(Cmd):

    def _decode(self, host, port, cluster_name, db_index, password=None, read_replica_host=None, read_replica_port=None):
        self.host = host
        self.port = port
        self.cluster_name = cluster_name
        self.db_index = db_index
        self.password = password
        self.read_replica_host = read_replica_host
        self.read_replica_port = read_replica_port

    def _execute(self):
        actor = self.actor()
        actor.log(
            "Connecting to %s port:%r, index:%r" % (self.host, self.port, self.db_index)
        )

        actor.host = self.host
        actor.port = self.port
        actor.db_index = self.db_index
        actor.password = self.password
        actor.read_replica_host = self.read_replica_host
        actor.read_replica_port = self.read_replica_port

        # We add those to environment so that subprocesses' session can use connect_from_env():
        os.environ.update(
            dict(
                KABARET_HOST=self.host,
                KABARET_PORT=self.port,
                KABARET_CLUSTER_NAME=self.cluster_name,
                KABARET_DB=self.db_index,
                KABARET_PASS=self.password or "",
                KABARET_READ_REPLICA_HOST=self.read_replica_host or "",
                KABARET_READ_REPLICA_PORT=self.read_replica_port or "",
            )
        )

        import redis.cluster, redis.sentinel

        # NOTE: decode_responses is needed with python 3 who wont accept bytes as string.

        db = redis.StrictRedis(
            host=self.host,
            port=self.port,
            password=self.password,
            decode_responses=True,
        )
        if self.read_replica_host is not None:
            actor.log(
                "Using replica to %s port:%r, index:%r" % (self.read_replica_host, self.read_replica_port, self.db_index)
            )

            read_db = redis.StrictRedis(
                host=self.read_replica_host,
                port=self.read_replica_port,
                password=self.password,
                decode_responses=True,
            )
            db = ReadWriteRedis(actor._session.log_debug, read_db, db)

        actor.set_db(
            db,
            self.cluster_name,
        )


@ClusterCmds.cmd
class Connect_From_Env(Cmd):
    """
    Will connect using the env var values.
    """

    def _decode(self):
        pass

    def _execute(self):
        actor = self.actor()
        actor.log("Connecting From Env...")
        try:
            host = os.environ["KABARET_HOST"]
            port = os.environ["KABARET_PORT"]
            cluster_name = os.environ["KABARET_CLUSTER_NAME"]
            db_index = os.environ["KABARET_DB"]
            password = os.environ.get("KABARET_PASS", None) or None
            read_replica_host = os.environ.get('KABARET_READ_REPLICA_HOST') or None
            read_replica_port = os.environ.get('KABARET_READ_REPLICA_PORT') or None
        except (KeyError, ValueError) as err:
            raise Exception(
                "Could not auto_connect the session (env value error): %s" % (err,)
            )
        else:
            actor.log(
                " => Env Connection: %s port:%r, index:%r" % (host, port, db_index)
            )

        actor.host = host
        actor.port = port
        actor.db_index = db_index
        actor.password = password
        actor.read_replica_host = read_replica_host
        actor.read_replica_port = read_replica_port

        # NOTE: decode_responses is needed with python 3 who wont accept bytes as string.
        db = redis.StrictRedis(
                host,
                port,
                db_index,
                password,
                decode_responses=True,
            )

        # This code is repeated from the Connect command :/
        # and it will be a problem on day.:
        if read_replica_host is not None:
            actor.log(
                " => Using replica to %s port:%r, index:%r" % (read_replica_host, read_replica_port, db_index)
            )

            read_db = redis.StrictRedis(
                read_replica_host,
                read_replica_port,
                db_index,
                password,
                decode_responses=True
            )
            db = ReadWriteRedis(actor._session.log_debug, read_db, db)
        
        actor.set_db(
            db, 
            cluster_name,
        )


@ClusterCmds.cmd
class Get_Connection_Info(Cmd):

    def _decode(self):
        pass

    def _execute(self):
        return self.actor().get_connection_info()


@ClusterCmds.cmd
class Disconnect(Cmd):

    def _decode(self):
        pass

    def _execute(self):
        self.actor().log("Disconnecting")


@ClusterCmds.cmd
class Broadcast(Cmd):

    def _decode(self, *words):
        self.words = words

    def _execute(self):
        self.actor().broadcast(*self.words)


@ClusterCmds.cmd
class Publish(Cmd):

    def _decode(self, **channels_messages):
        self.channels_messages = channels_messages

    def _execute(self):
        self.actor().publish(**self.channels_messages)


@ClusterCmds.cmd
class Get_Cluster_Name(Cmd):

    def _decode(self):
        pass

    def _execute(self):
        return self.actor().get_cluster_name()


class Cluster(Actor):

    def __init__(self, session):
        super(Cluster, self).__init__(session)

        self.KABARET_DB_INDEX = 3
        self.db = None
        self.pubsub = None

        self.host = None
        self.port = None
        self.db_index = None
        self.password = None
        self.cluster_name = None
        self.read_replica_host = None
        self.read_replica_port = None

        self._pubsubs = []

    def get_connection_info(self):
        """
        Returns the info used to connect.
        """
        return self.host, self.port, self.cluster_name, self.db_index

    def _create_cmds(self):
        return ClusterCmds(self)

    def _reg_pubsub(self, pubsub):
        self._pubsubs.append(pubsub)

    def _unreg_pubsub(self, pubsub, pattern):
        self._pubsubs.remove(pubsub)
        if pattern:
            pubsub.punsubscribe()
        else:
            pubsub.unsubscribe()

    def _read_pubsub(self):
        """
        Called periodically by the session since self.set_db() called.
        """
        # nb: we copy self._pubsubs so that handlers can unsubscribe
        # their channel (and thus modify self._pubsubs)

        if 0:
            # This was nice but it fetches only one message per pubsub.
            # If there's many messages, it will take many call to consume
            # every message :/
            for pubsub in list(self._pubsubs):
                unhandled_message = pubsub.get_message(
                    timeout=0  # dont wait for messages
                )
                if unhandled_message is not None:
                    self.on_unhandled_message(unhandled_message)

        else:
            # This is better since we consume messages until there's no
            # more.
            # Maybe we should make the maximum configurable ?
            #
            # NB: this code is derived from redis.client.PubSub.get_message()
            # It might be needed to update it if redis.client changes a lot
            # in further versions.
            for pubsub in list(self._pubsubs):
                max_iterations = 20
                curr = 0
                while curr < max_iterations:
                    response = pubsub.parse_response(block=False, timeout=0)
                    if response:
                        unhandled_message = pubsub.handle_message(
                            response, ignore_subscribe_messages=True
                        )
                        if unhandled_message:
                            self.on_unhandled_message(unhandled_message)
                    else:
                        curr = max_iterations

    def channels_subscribe(self, **channels_callbacks):
        """
        Returns a a callable used to unsuscribe all those subscriptions.
        """
        cluster_channels_callbacks = {}
        for ch, cb in channels_callbacks.items():
            channel = "%s:%s" % (self.cluster_name, ch)
            cluster_channels_callbacks[channel] = cb

        pubsub = self.db.pubsub(ignore_subscribe_messages=True)
        pubsub.subscribe(**cluster_channels_callbacks)
        self._reg_pubsub(pubsub)
        return lambda p=pubsub: self._unreg_pubsub(p, pattern=False)

    def patterns_subscribe(self, **patterns_callbacks):
        """
        Returns a a callable used to unsuscribe all those subscriptions.
        """
        cluster_patterns_callbacks = {}
        for pt, cb in patterns_callbacks.items():
            pattern = "%s:%s" % (self.cluster_name, pt)
            cluster_patterns_callbacks[pattern] = cb

        pubsub = self.db.pubsub(ignore_subscribe_messages=True)
        pubsub.psubscribe(**cluster_patterns_callbacks)
        self._reg_pubsub(pubsub)
        return lambda p=pubsub: self._unreg_pubsub(p, pattern=True)

    def broadcast(self, *words):
        message = " ".join([str(i) for i in words])
        self.publish(broadcast=message)

    def publish(self, **channels_messages):
        for channel, message in six.iteritems(channels_messages):
            cluster_channel = "%s:%s" % (self.cluster_name, channel)
            self.db.publish(cluster_channel, message)

    def on_unhandled_message(self, message):
        self.log("Unhandled Message:", repr(message))

    def on_broadcast_message(self, message):
        # message is a dict w/ pattern, type, channel, data
        self.log("[Broadcast Message]", repr(message["data"]))

        if 0:
            channels = self.db.pubsub_channels()
            self.log(
                "- Channels:",
                ", ".join(
                    ["%s=%i" % (c, i) for c, i in self.db.pubsub_numsub(*channels)]
                ),
            )
            self.log("- Patterns in use:", self.db.pubsub_numpat())

    def get_cluster_name(self):
        return self.cluster_name

    def get_db(self):
        if self.db is None:
            raise Exception("Cannot access the database without connection!")
        return self.db

    def set_db(self, db, cluster_name):
        if self.db is not None:
            raise Exception("The Database is already set !")

        try:
            db.ping()
        except redis.exceptions.ConnectionError as err:
            raise ValueError(
                str(err) + "\n\n" "=> Ask your administrator to check the database."
            )

        self.db = db
        self.db.client_setname(self.session().session_uid())

        self.cluster_name = cluster_name
        self.log("Connected to Cluster %r" % (self.cluster_name,))

        self.session().add_ticked(self._read_pubsub)
        self.session()._on_cluster_connected()

    def on_session_connected(self):
        self.broadcast_unsubscribe = self.channels_subscribe(
            broadcast=self.on_broadcast_message
        )
        self.broadcast("Cluster joined by", self.session().session_uid())

    def die(self):
        self.broadcast("Cluster left by", self.session().session_uid())
        self.broadcast_unsubscribe()



import os
import base64
from .interface import Editor_Interface

from six import text_type as unicode

from qtpy import QtWidgets, QtCore, QtGui
from kabaret.app import resources


class FileDialog(QtWidgets.QFileDialog):

    def __init__(self, options):
        QtWidgets.QFileDialog.__init__(self, caption="Select Image")

        # Force to select only one file
        self.setFileMode(self.FileMode.ExistingFile)
        # Set filters to images only
        self.setNameFilters(['Images (*.bmp *gif *.jpg *.png *.svg)'])
       
        # Select the current directory path
        if options.get('start_path'):
            self.setDirectory(options.get('start_path'))


class ImageViewer(QtWidgets.QWidget):
    '''
    This widget shows a preview of the image
    '''

    def __init__(self, pixmap=None):
        super().__init__()
        self.pixmap = None
        self.setPixmap(pixmap)

        self._sizeHint = QtCore.QSize()
        self.ratio = QtCore.Qt.KeepAspectRatio
        self.transformation = QtCore.Qt.SmoothTransformation

        self.setMaximumHeight(149)

    def setPixmap(self, pixmap):
        if self.pixmap != pixmap:
            self.pixmap = pixmap
            if isinstance(pixmap, QtGui.QPixmap):
                self._sizeHint = pixmap.size()
            else:
                self._sizeHint = QtCore.QSize()
            self.updateGeometry()
            self.updateScaled()

    def updateScaled(self):
        if self.pixmap:
            self.scaled = self.pixmap.scaled(self.size(), self.ratio, self.transformation)
        self.update()

    def sizeHint(self):
        return self._sizeHint

    def resizeEvent(self, event):
        self.updateScaled()

    def paintEvent(self, event):
        if not self.pixmap:
            return
        qp = QtGui.QPainter(self)
        r = self.scaled.rect()
        r.moveCenter(self.rect().center())
        qp.drawPixmap(r, self.scaled)


class ImageValueEditor(QtWidgets.QWidget, Editor_Interface):
    '''
    This editor lets you put a image using the file explorer or by drag and drop a file.

    Editor Type Names:
        image

    Options:
        start_path:      str - Open file explorer from a specific path
    '''

    @classmethod
    def can_edit(cls, editor_type_name):
        '''
        Must be implemented to return True if the given editor_type_name
        matches this editor.
        '''
        return editor_type_name in ('image',)

    def __init__(self, parent, options):
        QtWidgets.QWidget.__init__(self, parent)
        Editor_Interface.__init__(self, parent)

        self.options = options
        self.read_only = None

        self.setAcceptDrops(True)
        self.setFixedHeight(150)
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)

        self.build()
        self.apply_options(options)

    def build(self):
        # Draw a background for the editor
        self.frame = QtWidgets.QFrame()
        self.frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame.setObjectName('ImageEditor')

        # Layout when no image is setted
        self.input = QtWidgets.QWidget()
        input_layout = QtWidgets.QVBoxLayout()

        icon = QtGui.QIcon(resources.get_icon(('icons.gui', 'picture')))
        pixmap = icon.pixmap(QtCore.QSize(128, 128))
        icon_lbl = QtWidgets.QLabel('')
        icon_lbl.setPixmap(pixmap)
        
        self.label = QtWidgets.QLabel('Click or drop your image')

        input_layout.addWidget(icon_lbl, 0, QtCore.Qt.AlignCenter)
        input_layout.addWidget(self.label, 1, QtCore.Qt.AlignCenter)
        self.input.setLayout(input_layout)

        # Layout for ImageViewer
        preview_layout = QtWidgets.QVBoxLayout()
        preview_layout.setContentsMargins(5,5,5,5)
        self.preview = ImageViewer()
        preview_layout.addWidget(self.preview)

        # Buttons
        buttons_layout = QtWidgets.QVBoxLayout()
        buttons_layout.setContentsMargins(5,5,5,5)

        self.apply_button = QtWidgets.QToolButton()
        self.apply_button.setIcon(QtGui.QIcon(resources.get_icon(('icons.gui', 'correct-symbol'))))
        self.apply_button.setFixedSize(20,20)
        self.apply_button.setIconSize(QtCore.QSize(10,10))
        self.apply_button.clicked.connect(self._apply_edit)
        self.apply_button.setShortcut(QtGui.QKeySequence("Ctrl+Return"))

        self.clear_button = QtWidgets.QToolButton()
        self.clear_button.setIcon(QtGui.QIcon(resources.get_icon(('icons.gui', 'remove-symbol'))))
        self.clear_button.setFixedSize(20,20)
        self.clear_button.setIconSize(QtCore.QSize(10,10))
        self.clear_button.clicked.connect(self._on_clear_button_clicked)
        self.clear_button.setShortcut(QtGui.QKeySequence("Escape"))

        buttons_layout.addWidget(self.apply_button)
        buttons_layout.addStretch()
        buttons_layout.addWidget(self.clear_button)
        
        # Build main layout
        self.setLayout(QtWidgets.QGridLayout())
        self.layout().setContentsMargins(0,0,0,0)

        self.layout().addWidget(self.frame, 0, 0, 3, 0)
        self.layout().addWidget(self.input, 1, 0, QtCore.Qt.AlignCenter)
        self.layout().addLayout(preview_layout, 1, 0)
        self.layout().addLayout(buttons_layout, 0, 0, 3, 0, QtCore.Qt.AlignRight)

        self.preview.hide()
        self.apply_button.hide()
        self.clear_button.hide()

    def set_editable(self, b):
        '''
        Must be implemented to prevent editing if b is False.
        Visual cue show also be given to the user.
        '''
        self.read_only = not b
        if b is False:
            self.label.setText('No image')

    def apply_options(self, options):
        '''
        Must be implemented to configure the editor as
        described by the options dict.
        '''
        # Check value type for start_path option
        if options.get("start_path") and type(options.get("start_path")) is not str:
            raise Exception('start_path option should be setted with a str value.')

    def update(self):
        '''
        Must be implemnented to show the value returned by self.fetch_value()
        Your code should call self._on_updated() at the end.
        '''

        # Set current value on ImageViewer widget
        if self.fetch_value() is not None and unicode(self.fetch_value()):
            ba = QtCore.QByteArray.fromBase64(bytes(unicode(self.fetch_value()).split(',')[1], "utf-8"))
            pixmap = QtGui.QPixmap()
            pixmap.loadFromData(ba, unicode(self.fetch_value()).split(';')[0].split('/')[1])
            self.preview.setPixmap(pixmap)

            self.preview.show()
            if self.read_only is False:
                self.clear_button.show()
            self.apply_button.hide()
            self.input.hide()
        else:
            self.preview.setPixmap(None)
            self.preview.hide()
            self.clear_button.hide()
            self.apply_button.hide()
            self.input.show()

        self._on_updated()

    def get_edited_value(self):
        '''
        Must be implemented to return the value currently displayed.
        '''
        value = self.value
        if not value.isalpha():     # to avoid python keywords
            try:
                value = eval(value)
            except Exception:
                pass
        return value
 
    def _show_edited(self):
        '''
        Must be implemented to show that the displayed value
        needs to be applied.
        '''
        self.frame.setProperty('edited', True)
        self.frame.setProperty('applying', False)
        self.frame.style().polish(self.frame)

    def _show_applied(self):
        '''
        Must be implemented to show that the displayed value
        as been saved.
        In a clean scenario, applying edits will trigger an update()
        and this state should disapear.
        If you are using the Editor without this kind of round trip,
        you can call update here.
        '''
        self.frame.setProperty('applying', True)

    def _show_clean(self):
        '''
        Must be implemented to show that the displayed value is 
        up to date.
        '''
        self.frame.setProperty('edited', False)
        self.frame.setProperty('applying', False)
        self.frame.setProperty('error', False)
        self.frame.style().polish(self.frame)

    def _show_error(self, error_message):
        '''
        Must be implemented to show that the given error occured.
        '''
        self.frame.setProperty('error', True)
        self.frame.style().polish(self.frame)
        self.setToolTip('!!!\nERROR: %s' % (error_message,))

    def _on_clear_button_clicked(self):
        self.value = ''
        self.apply()

    def _apply_edit(self):
        self.apply()

    def setImage(self, path):
        # Encode image into base64 value
        with open(path, "rb") as image_file:
            encoded_string = base64.b64encode(image_file.read()).decode('utf-8')
            _, ext = os.path.splitext(path)
            self.value = f'data:image/{ext};base64,{encoded_string}'
        
        self.preview.setPixmap(QtGui.QPixmap(path))
        
        self.preview.show()
        self.clear_button.show()
        self.apply_button.show()
        self.input.hide()
        
        self._on_edited()

    def mousePressEvent(self, event):
        '''
        Open file explorer on mouse click
        '''
        if self.input.isVisible() and self.read_only is False:
            fd = FileDialog(self.options)
            if fd.exec_():
                paths = fd.selectedFiles()
                self.setImage(paths[0])
        
        super(ImageValueEditor, self).mousePressEvent(event)

    def dragEnterEvent(self, event):
        '''
        Only accept a image file on drag and drop
        '''
        if event.mimeData().hasUrls and self.read_only is False:
            if len(event.mimeData().urls()) == 1:
                formatCheck = QtGui.QImageReader.imageFormat(
                    event.mimeData().urls()[0].toLocalFile()
                )
                if formatCheck:
                    return event.accept()
        return event.ignore()

    def dropEvent(self, event):
        '''
        Set dropped image
        '''
        file_path = event.mimeData().urls()[0].toLocalFile()
        self.setImage(file_path)



from __future__ import print_function

import traceback

from qtpy import QtWidgets, QtCore

from ..styles import StylesManager

from ..icons import gui     # noqa # just declaring resources...
from ..icons import status  # noqa # just declaring resources...
from kabaret.app import resources


class DockTitleBar(QtWidgets.QWidget):
    def __init__(self, main_window_manager, dock, view):
        super(DockTitleBar, self).__init__(dock)
        self.mwm = main_window_manager
        self.dock = dock
        self.dock.topLevelChanged.connect(self.on_floating)
        self.view = view

        self.maximize_butt = None
        self.maximized = False
        self.installed = False

        try:
            self.view.set_on_view_title_change(self.on_view_title_change)
        except Exception as err:
            self.installed = False
            raise err
        else:
            try:
                self.install_tools()
            except Exception as err:
                self.installed = False
                raise err
            else:
                self.installed = True

    def sizeHint(self):
        return QtCore.QSize(5, 5)

    def minimumSizeHint(self):
        return QtCore.QSize(5, 5)

    def install_tools(self):
        self.maximize_butt = self.view.add_header_tool(
            'titleBarMaximize',
            self.maximized and '-' or '+',
            'Toggle Maximize',
            lambda self=self: self.mwm.toggle_maximized_dock(self)
        )
        self.installed = True

    def uninstall_tools(self):
        # FIXME: shouldn't a Hide be enough?
        self.view.remove_header_tool('titleBarMaximize')

    def on_view_title_change(self):
        self.dock.setWindowTitle(self.view.view_title())

    def on_floating(self, b):
        if b:
            self.dock.setTitleBarWidget(None)
            self.uninstall_tools()
        else:
            0 and self.dock.setTitleBarWidget(self)
            self.install_tools()
        self.dock.setFocus()

    def float_dock(self):
        if self.maximized:
            self.mwm.toggle_maximized_dock(self)
        self.dock.setFloating(True)

    def hide_dock(self):
        self.dock.hide()


class DockWidget(QtWidgets.QDockWidget):

    def __init__(self, dock_closed_callback, title, parent):
        super(DockWidget, self).__init__(title, parent)
        self.closed_callback = dock_closed_callback
        self.setFeatures(self.DockWidgetFeature.DockWidgetMovable | self.DockWidgetFeature.DockWidgetClosable)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)

    def closeEvent(self, e):
        self.closed_callback(self, self.view())

    def setFloating(self, floating):
        return False

    def view(self):
        return self.widget()

    def dock_area(self):
        mw = self.parentWidget()
        return mw.dockWidgetArea(self)


class MainWindow(QtWidgets.QMainWindow):

    def __init__(self, session, parent=None):
        super(MainWindow, self).__init__(parent)
        self.session = session

    def closeEvent(self, event):
        # Save views state on exit to be restored
        if self.session.layout_manager and self.session.layout_autosave:
            self.session.store_layout_preset(
                self.session.get_views_state(main_geometry=True),
                autosave=True
            )


class MainWindowManager(object):

    @classmethod
    def create_window(cls, session, parent=None):
        mw = MainWindow(session, parent)
        manager = cls(session, mw, embed_mode=False)
        return manager

    def __init__(self, session, main_window, embed_mode=False):
        super(MainWindowManager, self).__init__()
        self.main_window = main_window
        self.session = session
        self.embed_mode = embed_mode

    def install(self):
        mw = self.main_window

        if not self.embed_mode:
            mw.resize(QtCore.QSize(800, 800))
            StylesManager.get().apply_default_style()
            mw.setWindowIcon(
                resources.get_icon(('icons.gui', 'kabaret_icon'))
            )

            self.update_title()

            mw.setDockOptions(
                mw.DockOption.AnimatedDocks
                | mw.DockOption.AllowNestedDocks
                | mw.DockOption.AllowTabbedDocks
            )
            mw.setTabPosition(QtCore.Qt.AllDockWidgetAreas,
                              QtWidgets.QTabWidget.North)

        #!!self._view_types = {}
        self._docks = []
        self._maximised_to_restore = None

        self.session.register_view_types()
        mw.show()

    def update_title(self):
        title = self.session.session_name() or 'Kabaret Studio'
        title += '[%s]' % self.session.cmds.Cluster.get_cluster_name()
        self.main_window.setWindowTitle(title)

    def dock_closed(self, dock, view):
        if view.delete_on_close():
            view.about_to_delete()
            self._docks.remove(dock)
            dock.deleteLater()

        # TODO: Perhaps switch to restoring the focus to the previous widget?
        for dock in self._docks:
            view = dock.view()
            if view.view_type_name() == "Flow" and view.isVisible() is True:
                view.setFocus()
                break

    def dock_visibility_changed(self, visible, dock, view):
        if not visible:
            view.hidden = True
            view.on_hide()
        else:
            view.hidden = False
            view.on_show()
            if view.view_type_name() == "Flow" and self.session.current_view() is not view:
                view.setFocus()

    def dock_location_changed(self, location, dock, view):
        view.area = location

    def register_view_type(self, Type):
        print('!!! WARNING !!!')
        print('!!! WARNING !!! session.main_window_manager.register_view_type() is deprecated !')
        print('!!! WARNING !!! (You must use session.register_view_type() instead)')
        print('!!! WARNING !!!')
        return self.session.register_view_type(Type)

    def add_view(self, type_name, area=None, visible=True, *view_args, **view_kwargs):
        print('!!! WARNING !!!')
        print('!!! WARNING !!! session.main_window_manager.add_view() is deprecated !')
        print('!!! WARNING !!! (You must use session.add_view() instead)')
        print('!!! WARNING !!!')
        if area is not None:
            print('!!! WARNING !!!')
            print('!!! WARNING !!! session.main_window_manager.add_view() is deprecated !')
            print('!!! WARNING !!! The "area" kword is not supported anymore')
            print('!!! WARNING !!!')
        if area is not None:
            print('!!! WARNING !!!')
            print('!!! WARNING !!! session.main_window_manager.add_view() is deprecated !')
            print('!!! WARNING !!! The "visible" kword is not supported anymore')
            print('!!! WARNING !!!')

        return self.session.add_view(type_name, *view_args, **view_kwargs)

    def add_toolbar_view(self, view, hidden=False, area=None):
        area = area or QtCore.Qt.TopToolBarArea
        self.main_window.addToolBar(area, view)
        view.setVisible(not hidden)

    def create_docked_view_dock(self, view, hidden=False, area=None):
        dock = DockWidget(self.dock_closed, view.view_title(), self.main_window)
        tb = DockTitleBar(self, dock, view)
        if tb.installed:
            0 and dock.setTitleBarWidget(tb)
        else:
            tb.deleteLater()

        dock.setWidget(view)
        dock.visibilityChanged.connect(
            lambda visible, dock=dock, view=view: self.dock_visibility_changed(visible, dock, view)
        )
        dock.dockLocationChanged.connect(
            lambda area, dock=dock, view=view: self.dock_location_changed(area, dock, view)
        )
        dock.setObjectName('Dock_%s_%i' % (view.view_title(), len(self._docks),))

        area = area or QtCore.Qt.LeftDockWidgetArea
        self.main_window.addDockWidget(area, dock)

        if hidden:
            dock.hide()

        # Auto tabify views of matching type name:
        target_view = self.find_docked_view(view.view_type_name(), area=area)
        if target_view is not None:
            target_view.dock_widget().show()
            dock.show()
            self.main_window.tabifyDockWidget(target_view.dock_widget(), dock)

        self._docks.append(dock)

        if len(self._docks) > 1:
            for d in self._docks:
                if d.titleBarWidget() and d.titleBarWidget().maximize_butt:
                    d.titleBarWidget().maximize_butt.setEnabled(True)
        else:
            for d in self._docks:
                if d.titleBarWidget() and d.titleBarWidget().maximize_butt:
                    d.titleBarWidget().maximize_butt.setEnabled(False)

        return dock

    def tabify_view(self, first_view, second_view):
        d1 = first_view.dock_widget()
        d2 = second_view.dock_widget()
        self.main_window.tabifyDockWidget(d1, d2)

    def toggle_maximized_dock(self, titlebar):
        button = titlebar.maximize_butt
        keeped = titlebar.dock

        if self._maximised_to_restore is None:
            button.setText('-')
            self._maximised_to_restore = self.main_window.saveState()
            for d in self._docks:
                if d is keeped:
                    d.setFocus()
                    continue
                d.hide()

        else:
            button.setText('+')

            self.main_window.restoreState(self._maximised_to_restore)
            self._maximised_to_restore = None

    def find_docked_view(self, view_type_name, area=None):
        for dock in self._docks:
            if area is not None and dock.dock_area() != area:
                continue
            view = dock.view()
            if view.view_type_name() == view_type_name:
                return view
        return None

    def find_view(self, view_type_name, create=False, area=None, *args, **kwargs):
        '''
        Returns a view with the specified type.
        If create is True and no existsing view in found, create one in the given area
        using *args and **kwargs.
        If create is False, None is returned.
        '''
        print('!!! WARNING !!!')
        print('!!! WARNING !!! session.main_window_manager.find_view() is deprecated !')
        print('!!! WARNING !!! (You must use session.find_view() instead)')
        print('!!! WARNING !!!')
        if area is not None:
            print('!!! WARNING !!!')
            print('!!! WARNING !!! session.main_window_manager.find_view() is deprecated !')
            print('!!! WARNING !!! The "area" kword is not supported anymore')
            print('!!! WARNING !!!')
        return self.session.find_view(view_type_name, create=create, *create_args, **create_kwargs)
        # try:
        #     Type = self._view_types[view_type_name]
        # except KeyError:
        #     raise ValueError('Find View: Unknown view type %r' %
        #                      (view_type_name,))

        # for dock in self._docks:
        #     if area is not None and dock.dock_area() != area:
        #         continue
        #     view = dock.view()
        #     if view.__class__ == Type:
        #         return view

        # if create:
        #     return self.add_view(view_type_name, area, *args, **kwargs)
        # else:
        #     return None

    # def receive_event(self, event_type, data):
    #     pass
    #     if event_type == 'cluster_connection':
    #         cluster_name = data['cluster_name']
    #         self._set_title(cluster_name)

    # def dispatch_event(self, event_type, data):
    #     if event_type == 'cluster_connection':
    #         cluster_name = data['cluster_name']
    #         self._set_title(cluster_name)

    #     for T in self._view_types.values():
    #         for view in T._VIEWS:
    #             view.receive_event(event_type, data)

    def resize(self, width, height):
        self.main_window.resize(width, height)

    def get_geometry_state(self):
        return self.main_window.saveGeometry()
    
    def set_geometry_state(self, state):
        self.main_window.restoreGeometry(state)
    
    def get_layout_state(self):
        return self.main_window.saveState()

    def set_layout_state(self, state):
        self.main_window.restoreState(state)




import os
import six
import getpass
import functools

from qtpy import QtWidgets, QtCore, QtGui
from datetime import datetime

from .widget_view import ToolBarView
from kabaret.app import resources


LAYOUTS = {}

class ClickLabel(QtWidgets.QLabel):

    def __init__(self, on_click, *args, **kwargs):
        super(ClickLabel, self).__init__(*args, **kwargs)
        self._on_click = on_click

    def mousePressEvent(self, event):
        self._on_click()

class SessionToolBar(ToolBarView):

    def __init__(self, *args, **kwargs):
        super(SessionToolBar, self).__init__(*args, **kwargs)

        self.user_label = QtWidgets.QLabel(self)
        self.session_id_label = QtWidgets.QLabel(self)

        stretch = QtWidgets.QWidget(self)
        stretch.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding,
            QtWidgets.QSizePolicy.Preferred,
        )

        self.addWidget(self.user_label)
        self.addSeparator()
        self.addWidget(self.session_id_label)
        self.addSeparator()

        # Show layout menu when manager is enabled
        if self.session.layout_manager:
            self.layouts_tb = QtWidgets.QToolButton(self)
            self.layouts_tb.setText('Layouts')
            self.layouts_tb.setPopupMode(self.layouts_tb.ToolButtonPopupMode.InstantPopup)
            self.layouts_tb.setIcon(resources.get_icon(('icons.gui','ui-layout')))
            self.layouts_tb.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            self.layout_menu = QtWidgets.QMenu(self.layouts_tb)
            self.layout_menu.addAction('TEST')
            self.layout_menu.aboutToShow.connect(self.update_layout_tb)
            self.layouts_tb.setMenu(self.layout_menu)

            self.addWidget(stretch)
            self.addSeparator()
            self.addWidget(self.layouts_tb)
            
        self.update_user_label()

    def update_user_label(self):
        label = ' <b><font size=+5>{}</font></b> '.format(
            getpass.getuser(),
        )
        self.user_label.setText(label)

        label=' {}[{}] '.format(
            self.session.session_name(),
            self.session.cmds.Cluster.get_cluster_name(),
        )
        self.session_id_label.setText(label)
        
        self.session_id_label.setToolTip(
            '<b>Session ID:</b> {}<br><b>Connection:</b> {}'.format(
                self.session.session_uid(),
                ' - '.join(self.session.cmds.Cluster.get_connection_info())
            )
        )

    def update_layout_tb(self):
        # Clear all actions of context menu
        tb = self.layouts_tb
        menu = tb.menu()
        menu.clear()

        # Split layouts by project
        project_menus = {}
        projects_info = self.session.get_actor("Flow").get_projects_info()
        for project in projects_info:
            name = project[0]
            project_menus[name] = menu.addMenu(name)

        # Add user layout presets
        self.layout_icon = resources.get_icon(('icons.gui', 'ui-layout'))
        for name, layout in sorted(six.iteritems(self.session.get_layout_presets())):
            project_found = False
            
            for project_name, project_menu in project_menus.items():
                for view_data in layout['views']:
                    view_state = view_data[-1]
                    if 'oid' in view_state:
                        if view_state['oid'].startswith(f'/{project_name}'):
                            action = QtGui.QAction(self.layout_icon, name, project_menu)
                            action.triggered.connect(functools.partial(self._on_set_layout_action, layout))
                            action.setObjectName('la')

                            project_menu.addAction(action)
                            project_found = True
                            break

            # Add directly to layout root menu if no project has been found in views
            if not project_found:
                action = QtGui.QAction(self.layout_icon, name)
                action.triggered.connect(functools.partial(self._on_set_layout_action, layout))
                action.setObjectName('la')

                menu.insertAction(project_menu.menuAction(), action)
        
        # Add a separator
        self.menu_sep = QtGui.QAction(menu)
        self.menu_sep.setSeparator(True)
        self.menu_sep.setObjectName('sep')
        menu.addAction(self.menu_sep)

        # Add store layout action
        icon = resources.get_icon(('icons.gui', 'plus-symbol-in-a-rounded-black-square'))
        store_action = QtGui.QAction(icon, 'Store Current Layout', menu)
        store_action.triggered.connect(self._on_store_current_layout_action)
        store_action.setObjectName('sa')
        menu.addAction(store_action)

        # Add delete layout action
        icon = resources.get_icon(('icons.gui', 'minus-button'))
        delete_action = QtGui.QAction(icon, 'Delete Layout', menu)
        delete_action.triggered.connect(self._on_delete_layout_action)
        delete_action.setObjectName('da')
        menu.addAction(delete_action)

        # Add layout session autosaves
        icon = resources.get_icon(('icons.gui', 'share-post-symbol'))
        recover_menu = menu.addMenu(icon, 'Recover Session Layout')

        for name, layout in sorted(six.iteritems(self.session.get_layout_presets(autosaves=True)), reverse=True):
            action = QtGui.QAction(self.layout_icon, name, self.layouts_tb.menu())
            action.triggered.connect(functools.partial(self._on_set_layout_action, layout))
            action.setObjectName('la')

            recover_menu.addAction(action)

    def _on_set_layout_action(self, layout):
        # Don't delete us inside an signal handler from us:
        QtCore.QTimer.singleShot(
            100, lambda l=layout, s=self.session: s.set_views_state(l)
        )

    def _on_store_current_layout_action(self):
        # Create dialog
        dialog = QtWidgets.QDialog(self)
        dialog.setWindowTitle('Save Layout')
        dialog.setLayout(QtWidgets.QVBoxLayout())

        # Create title
        title = QtWidgets.QLabel('<h1>Enter preset name</h1>')
        dialog.layout().addWidget(title)

        # Create combobox selection
        cb = QtWidgets.QComboBox(dialog)
        cb.addItems(['']+list(self.session.get_layout_presets().keys()))
        cb.setEditable(True)
        cb.lineEdit().setValidator(QtGui.QRegularExpressionValidator(QtCore.QRegularExpression('[^<>:"\/\\\|?*]+')))
        dialog.layout().addWidget(cb)

        # Create checkbox for window position
        checkbox = QtWidgets.QCheckBox("Save Window Position")
        checkbox.setChecked(False)
        dialog.layout().addWidget(checkbox)

        # Create button
        b = QtWidgets.QPushButton(dialog)
        b.setText('Save Layout')
        b.clicked.connect(dialog.accept)
        dialog.layout().addWidget(b)

        dialog.setFixedSize(350, 150)

        # Store and update menu if accepted
        cancel = dialog.exec_() != dialog.Accepted
        name = cb.currentText().strip()
        dialog.deleteLater()
        if cancel or not name:
            return

        self.session.store_layout_preset(
            self.session.get_views_state(main_geometry=checkbox.isChecked()),
            name
        )
        self.update_layout_tb()

    def _on_delete_layout_action(self):
        # Create dialog
        dialog = QtWidgets.QDialog(self)
        dialog.setWindowTitle('Delete Layout')
        dialog.setLayout(QtWidgets.QVBoxLayout())

        # Create title
        title = QtWidgets.QLabel('<h1>Select layouts to delete</h1>')
        dialog.layout().addWidget(title)

        # Create user layout presets list
        listw = QtWidgets.QListWidget(dialog)
        for layout_name in list(self.session.get_layout_presets().keys()):
            item = QtWidgets.QListWidgetItem(layout_name)
            item.setCheckState(QtCore.Qt.Unchecked)
            listw.addItem(item)
        dialog.layout().addWidget(listw)

        # Create button
        b = QtWidgets.QPushButton(dialog)
        b.setText('Delete')
        b.clicked.connect(dialog.accept)
        dialog.layout().addWidget(b)

        dialog.setFixedSize(375, 300)

        # Delete and update menu if accepted
        cancel = dialog.exec_() != dialog.Accepted
        dialog.deleteLater()
        if cancel:
            return

        checked = []
        for row in range(listw.count()):
            item = listw.item(row)
            if item.checkState():
                checked.append(item.text())

        self.session.delete_layout_preset(checked)
        self.update_layout_tb()

    def receive_event(self, event_type, data):
        pass



'''
    The Flow let you model you project structure, pipeline and tracking strategies.

    < insert introductionnal description here >
'''

WIP = '''


    Changes to implement:

    - An object can holds override info for each of its relation ui_config (icon, editor, hidden, enabled, ...)
    - the ui_config must give info to have several relation on the same line.
    - use a context to declare relations into groups
    - use a context to declare relations into line
    - use typed value in param relation, with type validation (and leverage redis value types)
    - object summary must be a related object (with touched events emited when changed)
        -> gui finds it by relation name '_summary'
    - there must be an option to the computed relation to cache the result between touch() calls.

    - Refs stored as set in redis
    - Have indexed flag on values (using Value.get_index_oid(), configurable in relation)
    - Have inherited flag on values (default is provided by Value.get_inherited_value(), configuragle in relation)
    - Map -> Hash, use a special type in _items to leverage redis Hash value (speed + stable default ordering)
    - Map should store a 'is_hidden' bool alongside the full_qualified_type_name to allow filtering of
        mapped item w/o instanciating all of them (which is required if the info is store in a value of items)
    - Drop Map row_icons, use a special column '_style' in rows that olds icon, colors, etc...
    - Map.rows() must accept a page len and a page number

    - value store dont send old/new values on set methods (network noise for nothing)
    - object should not use parent's value store but the return value of parent.default_related_value_store
        (so Object can override its value store and still provide default value store for children)
    - computed values must not be saved in the value store (maybe add an option to do it sometimes...)

    - relations to values must have an option "dont_store_value" so object can have some on its
    related value in memory only (and no all related object like when you override the value store)
    - in memory value should not send touch event to other sessions

    - Action._buttons => this sucks, find something better.
    - Action must be able to be visible in parent inline w/o being visible in parent page
    - DropAction MUST BE ABLE TO HANDLE SEVERAL SOURCES.

    - drop the dirty flags for now and see how it goes.

    - see if Object instanciation is too costly (__new__ stuff), and fix it if needed.

    - Connection (Refs...) are only used for flow configuration by the user (casting, todo list...)
        when an Ojbect needs to know another -not directly related one (blast's maya scene),
        it should use a method returning that object (so that subclasses can override it)
        -> drop the smoky "PipelineConnection" stuff

    Wish list:
    - Action proxy (a relation to a relative action, so that it is promoted to (grand)parents )
    - Action chaining (aka wizards)
    - A concept of Next / Prev on objects.
        -> if parent is a map, navigate the map
        -> return prev/next related object, or parent's prev last relation / next first relation.
        -> parent must be able to override the behavior.

'''

from .exceptions import (
    MissingChildError, MissingRelationError, MappedNameError,
    RefSourceTypeError, RefSourceError,
)

from .relations import (
    Parent, Child, Relative,
    Separator, Label,
    Param, SessionParam,
    IntParam, BoolParam, FloatParam, StringParam,
    DictParam, OrderedStringSetParam, HashParam,
    Computed,
    Connection
)
from . import values
from .value_store import AbstractValueStore, MemoryValueStore

from .object import Object, group, SessionObject, Root, ThumbnailInfo
from .action import (
    Action, ConnectAction,
    ChoiceValueSetAction, ChoiceValueSelectAction,
)
from .map import DynamicMap, Map, RefsMap
MAP_TYPES = (DynamicMap, Map)

from .injection import injection, InjectionProvider